import os
import re
from os.path import basename, dirname, exists, join, normpath, pardir
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import nrrd
import numpy as np
import pydicom
from pydicom import FileDataset
from tqdm import tqdm

from supervisely import image, logger, volume
from supervisely.annotation.tag import Tag
from supervisely.io.fs import (
    dir_exists,
    get_file_ext,
    get_file_name,
    get_file_name_with_ext,
    mkdir,
)

_MEDICAL_DEFAULT_GROUP_TAG_NAMES = [
    "StudyInstanceUID",
    "StudyID",
    "SeriesInstanceUID",
    "TreatmentSessionUID",
    "Manufacturer",
    "ManufacturerModelName",
    "Modality",
]


def slice_nifti_file(nii_file_path: str, converted_dir: str) -> Tuple[List[str], List[str]]:
    """Slices file into 2D slices if it is 3D.
    Returns image paths and image names.

    This function always slices 3D nifti files into 2D slices by axial orientation because source data automatically gets converted to RAS coordinate system.
    """

    output_name = get_file_name(nii_file_path)
    if get_file_ext(output_name) == ".nii":
        output_name = get_file_name(output_name)

    data, header = volume.convert_nifti_to_nrrd(nii_file_path)

    output_paths = []
    output_names = []
    if len(data.shape) >= 3:
        indices = [slice(None)] * 3 + [0] * (data.ndim - 3)

        for i in range(data.shape[2]):
            indices[2] = i
            slice_data = data[tuple(indices)]

            new_header = {
                "sizes": slice_data.shape,
                "type": header.get("type", "float"),
                "dimension": len(slice_data.shape),
            }
            output_path = os.path.join(converted_dir, f"nifti_{output_name}_{i}.nrrd")
            output_names.append(get_file_name_with_ext(output_path))
            nrrd.write(output_path, slice_data, new_header)
            output_paths.append(output_path)
    else:
        output_path = os.path.join(converted_dir, f"nifti_{output_name}.nrrd")
        output_names.append(get_file_name_with_ext(output_path))
        nrrd.write(output_path, data, header)
        output_paths.append(output_path)

    return output_paths, output_names


def is_dicom_file(path: str) -> bool:
    """Checks if file is dicom file by given path."""
    try:
        pydicom.read_file(str(Path(path).resolve()), stop_before_pixels=True)
        result = True
    except Exception as ex:
        logger.warning("'{}' appears not to be a DICOM file\n({})".format(path, ex))
        result = False
    return result


def check_nrrd(path):
    try:
        nrrd.read(path)
        return True
    except:
        return False


def create_nrrd_header_from_dcm(dcm_path: str, frame_axis: int = 2) -> dict:
    """
    Create nrrd header from DICOM file.

    Assume that axes will be in the RAS order (right-anterior-superior).
    """
    _, meta = volume.read_dicom_serie_volume([dcm_path], False)
    dimensions: Dict = meta.get("dimensionsIJK")
    header = {
        "type": "float",
        "sizes": [dimensions.get("x"), dimensions.get("y")],
        "dimension": 2,
        "space": "right-anterior-superior",
    }

    if frame_axis == 0:
        spacing = meta["spacing"][1:]
        header["space directions"] = [[spacing[0], 0], [0, spacing[1]]]
    if frame_axis == 1:
        spacing = meta["spacing"][0::2]
        header["space directions"] = [[spacing[0], 0], [0, spacing[1]]]
    if frame_axis == 2:
        spacing = meta["spacing"][0:2]
        header["space directions"] = [[spacing[0], 0], [0, spacing[1]]]
    return header


def find_frame_axis(pixel_data: np.ndarray, frames: int) -> int:
    for axis in range(len(pixel_data.shape)):
        if pixel_data.shape[axis] == frames:
            return axis
    raise ValueError("Unable to recognize the frame axis for splitting a set of images")


def create_pixel_data_set(dcm: FileDataset, frame_axis: int) -> Tuple[List[np.ndarray], int]:
    if frame_axis == 0:
        pixel_array = np.transpose(dcm.pixel_array, (2, 1, 0))
    elif frame_axis == 1:
        pixel_array = np.transpose(dcm.pixel_array, (2, 0, 1))
    else:
        pixel_array = dcm.pixel_array
    frame_axis = 2
    list_of_images = np.split(pixel_array, int(dcm.NumberOfFrames), axis=frame_axis)
    return list_of_images, frame_axis

def convert_to_monochrome2(dcm_path: str, dcm: FileDataset) -> FileDataset:
    if getattr(dcm, "PhotometricInterpretation", None) == "YBR_FULL_422":
        # * Convert dicom to monochrome
        if len(dcm.pixel_array.shape) == 4 and dcm.pixel_array.shape[-1] == 3:
            monochrome = dcm.pixel_array[..., 0].astype(np.uint8)
        else:
            logger.warn("Unexpected shape for YBR_FULL_422 data: " + str(dcm.pixel_array.shape))

        try:
            dcm.SamplesPerPixel = 1
            dcm.PhotometricInterpretation = "MONOCHROME2"
            dcm.PlanarConfiguration = 0
            if len(monochrome.shape) == 3:
                dcm.NumberOfFrames = str(monochrome.shape[0])
                dcm.Rows, dcm.Columns = monochrome.shape[1:3]
            dcm.PixelData = monochrome.tobytes()
        except AttributeError as ae:
            logger.error(f"Error occurred while converting dicom to monochrome: {ae}")

        logger.info("Rewriting DICOM file with monochrome2 format")
        dcm.save_as(dcm_path)
    return dcm

def convert_dcm_to_nrrd(
    image_path: str, converted_dir: str, group_tag_name: Optional[list] = None
) -> Tuple[List[str], List[str], List[dict], dict]:
    """Converts DICOM data to nrrd format and returns image paths and image names"""
    original_name = get_file_name_with_ext(image_path)
    if not dir_exists(converted_dir):
        mkdir(converted_dir)
    curr_convert_dir = os.path.join(converted_dir, original_name)
    mkdir(curr_convert_dir)

    dcm = pydicom.read_file(image_path)
    try:
        if dcm.file_meta.TransferSyntaxUID.is_compressed:
            dcm.decompress()
    except AttributeError:
        logger.warn("Couldn't find key 'TransferSyntaxUID' in dicom's metadata.")
    dcm = convert_to_monochrome2(image_path, dcm)

    dcm_meta = get_dcm_meta(dcm)

    if group_tag_name is None:
        group_tag_names = _MEDICAL_DEFAULT_GROUP_TAG_NAMES
    else:
        group_tag_names = [group_tag_name] + _MEDICAL_DEFAULT_GROUP_TAG_NAMES

    is_group_tag_exists = next((name for name in group_tag_names if name in dcm), None) is not None
    if not is_group_tag_exists:
        logger.warning(f"Not found group tags in DICOM file.")

    header, frames, frame_axis, pixel_data_list = get_nrrd_data(image_path, dcm)

    group_tags = []
    save_paths = []
    image_names = []
    frames_list = [f"{i:0{len(str(frames))}d}" for i in range(1, frames + 1)]

    for pixel_data, frame_number in zip(pixel_data_list, frames_list):
        if frames == 1:
            pixel_data = image.rotate(img=pixel_data, degrees_angle=270)
            pixel_data = image.fliplr(pixel_data)
            image_name = f"{original_name}.nrrd"
        else:
            pixel_data = np.squeeze(pixel_data, frame_axis)
            image_name = f"{frame_number}_{original_name}.nrrd"

        save_path = os.path.join(curr_convert_dir, image_name)
        nrrd.write(save_path, pixel_data, header)
        save_paths.append(save_path)
        image_names.append(image_name)

    for tag_name in group_tag_names:
        if dcm.get(tag_name) is not None:
            group_tag_value = dcm[tag_name].value
            if group_tag_value is not None and group_tag_value != "":
                group_tags.append({"name": tag_name, "value": group_tag_value})
        elif group_tag_name is not None and tag_name == group_tag_name:
            logger.warning(f"Couldn't find key {group_tag_name!r} in file's metadata.")

    return save_paths, image_names, group_tags, dcm_meta


def slice_nrrd_file(nrrd_file_path: str, output_dir: str) -> Tuple[List[str], List[str]]:
    """Slices nrrd file into 2D slices if it is 3D. Returns image paths and image names.

    This function always slices 3D nrrd files into 2D slices by axial orientation because source data automatically gets converted to RAS coordinate system.
    Data with dimensions more than 3 will be truncated to 3D.
    """
    data, header = nrrd.read(nrrd_file_path)
    output_paths = []
    output_names = []

    if header.get("dimension", 0) > 2:
        logger.info(f"File [{nrrd_file_path}] have more than 2 dimensions.")
        try:
            domain_count = 0
            if "kinds" in header:
                kinds = header.get("kinds")
                domain_indices = [i for i, kind in enumerate(kinds) if kind == "domain"]
                domain_count = len(domain_indices)
            else:
                domain_indices = [0, 1, 2]
                domain_count = 3

            # If there are more than three domain dimensions, ignore the extra ones
            domain_indices = domain_indices[:3]

            # Create a new array with only the first three domain dimensions
            indices = [0] * data.ndim
            for idx in domain_indices:
                indices[idx] = slice(None)
            data = data[tuple(indices)]
            domain_indices = [0, 1, 2]

            # If shape is 3D and space is not RAS, then we need to transpose the data
            if domain_count == 3 and "space" in header:
                if header.get("space").lower() not in ["right-anterior-superior", "ras"]:
                    domain_indices = domain_indices[::-1]
                    data = np.transpose(data, domain_indices)

            progress = tqdm(range(data.shape[-1]), desc="Slicing images")

            for i in range(data.shape[-1]):
                slice_data = np.take(data, i, axis=len(data.shape) - 1)

                new_header = {
                    "sizes": slice_data.shape,
                    "type": header.get("type", "float"),
                    "dimension": len(slice_data.shape),
                }

                output_nrrd_path = os.path.join(
                    output_dir, f"{os.path.basename(nrrd_file_path).replace('.nrrd', '')}_{i}.nrrd"
                )
                output_name = get_file_name_with_ext(output_nrrd_path)
                nrrd.write(output_nrrd_path, slice_data, new_header)
                output_paths.append(output_nrrd_path)
                output_names.append(output_name)
                progress.update(1)
            progress.close()
        except Exception:
            logger.warn(f"File [{nrrd_file_path}] is not supported. Skipping...")
    else:
        output_paths.append(nrrd_file_path)
        output_names.append(get_file_name_with_ext(nrrd_file_path))
    return output_paths, output_names


def get_dcm_meta(dcm: FileDataset) -> List[Tag]:
    """Create tags from DICOM metadata."""
    from supervisely.volume.volume import _anonymize_tags

    filtered_tags = []

    filename = get_file_name_with_ext(dcm.filename)
    empty_tags, too_long_tags = [], []
    for dcm_tag in dcm.keys():
        try:
            curr_tag = dcm[dcm_tag]
            dcm_tag_name = str(curr_tag.name)
            dcm_tag_value = str(curr_tag.value)
            if dcm_tag_name in ["Patient's Name", "Patient ID"] + _anonymize_tags:
                dcm_tag_value = "anonymized"
            if dcm_tag_value in ["", None]:
                empty_tags.append(dcm_tag_name)
                continue
            if len(dcm_tag_value) > 255:
                too_long_tags.append(dcm_tag_name)
                continue
            filtered_tags.append((dcm_tag_name, dcm_tag_value))
        except KeyError:
            dcm_filename = get_file_name_with_ext(dcm.filename)
            logger.warning(f"Couldn't find key '{dcm_tag}' in file's metadata: '{dcm_filename}'")
            continue

    if len(empty_tags) > 0:
        logger.warning(f"{filename}: {len(dcm_tag_name)} tags have empty value. Skipped tags: {empty_tags}.")
    if len(too_long_tags) > 0:
        logger.warning(
            f"{filename}: {len(too_long_tags)} tags have too long value (> 255 symbols). Skipped tags: {too_long_tags}."
        )

    dcm_tags_dict = {}
    for dcm_tag_name, dcm_tag_value in filtered_tags:
        dcm_tags_dict[dcm_tag_name] = dcm_tag_value
    return dcm_tags_dict


def get_nrrd_data(image_path: str, dcm: FileDataset):

    frame_axis = 2
    pixel_data_list = [dcm.pixel_array]

    if len(dcm.pixel_array.shape) == 3:
        if dcm.pixel_array.shape[0] == 1 and not hasattr(dcm, "NumberOfFrames"):
            frames = 1
            pixel_data_list = [
                dcm.pixel_array.reshape((dcm.pixel_array.shape[1], dcm.pixel_array.shape[2]))
            ]
            header = create_nrrd_header_from_dcm(image_path, frame_axis)
        else:
            try:
                frames = int(dcm.NumberOfFrames)
            except AttributeError as e:
                if str(e) == "'FileDataset' object has no attribute 'NumberOfFrames'":
                    e.args = ("can't get 'NumberOfFrames' from dcm meta.",)
                    raise e
            frame_axis = find_frame_axis(dcm.pixel_array, frames)
            pixel_data_list, frame_axis = create_pixel_data_set(dcm, frame_axis)
            header = create_nrrd_header_from_dcm(image_path, frame_axis)
    elif len(dcm.pixel_array.shape) == 2:
        frames = 1
        header = create_nrrd_header_from_dcm(image_path, frame_axis)
    else:
        raise RuntimeError(
            f"This type of dcm data is not supported, pixel_array.shape = {len(dcm.pixel_array.shape)}"
        )

    return header, frames, frame_axis, pixel_data_list



import os
from pathlib import Path
from typing import Dict, Optional, Union

import supervisely.convert.image.cityscapes.cityscapes_helper as helper
from supervisely import (
    Annotation,
    ObjClass,
    ObjClassCollection,
    Polygon,
    ProjectMeta,
    Tag,
    TagMeta,
    TagValueType,
    logger,
)
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.imaging.color import generate_rgb
from supervisely.io.fs import JUNK_FILES, get_file_name
from supervisely.io.json import load_json_file
from supervisely.project.project_settings import LabelingInterface


class CityscapesConverter(ImageConverter):

    def __init__(
            self,
            input_data: str,
            labeling_interface: Optional[Union[LabelingInterface, str]],
            upload_as_links: bool,
            remote_files_map: Optional[Dict[str, str]] = None,
    ):
        super().__init__(input_data, labeling_interface, upload_as_links, remote_files_map)

        self._classes_mapping = {}
        self._supports_links = True
        self._force_shape_for_links = self.upload_as_links

    def __str__(self):
        return AvailableImageConverters.CITYSCAPES

    @property
    def key_file_ext(self) -> str:
        return ".json"
    
    @property
    def ann_ext(self) -> str:
        return ".json"

    def ann_file_ext(self) -> str:
        return ".json"

    def validate_ann_file(self, ann_file_path: str) -> bool:
        try:
            ann_json = load_json_file(ann_file_path)
            if isinstance(ann_json, dict):
                objects = ann_json.get("objects")
                if objects is not None:
                    return True
                else:
                    logger.warn(f"Couldn't read objects from annoation file: '{ann_file_path}'")
                    return False
            else:
                return False
        except:
            logger.warn(f"Failed to read annotation file: '{ann_file_path}'")
            return False

    def validate_key_file(self, key_file_path: str) -> bool:
        try:
            classes_mapping = load_json_file(key_file_path)
            if isinstance(classes_mapping, list):
                obj_classes = []
                for obj_class_json in classes_mapping:
                    if (
                        obj_class_json.get(
                            "name",
                        )
                        is None
                    ):
                        continue
                    obj_class_name = obj_class_json["name"]
                    if helper.CITYSCAPES_CLASSES_TO_COLORS_MAP.get(obj_class_name, None):
                        obj_class = ObjClass(
                            name=obj_class_name,
                            geometry_type=Polygon,
                            color=helper.CITYSCAPES_CLASSES_TO_COLORS_MAP[obj_class_name],
                        )
                    else:
                        if obj_class_json.get("color") is None:
                            new_color = obj_class_json["color"]
                        else:
                            new_color = generate_rgb(helper.CITYSCAPES_COLORS)
                        helper.CITYSCAPES_COLORS.append(new_color)
                        obj_class = ObjClass(
                            name=obj_class_name, geometry_type=Polygon, color=new_color
                        )
                    obj_classes.append(obj_class)
                obj_class_collection = ObjClassCollection(obj_classes)
                self._classes_mapping = classes_mapping
                tag_meta = TagMeta("split", TagValueType.ANY_STRING)
                if self._meta is None:
                    self._meta = ProjectMeta(obj_classes=obj_class_collection, tag_metas=[tag_meta])
                else:
                    if not self._meta.tag_metas.has_key("split"):
                        self._meta = self._meta.add_tag_meta(tag_meta)
                    self._meta = self._meta.add_obj_classes(obj_classes)
                return True
            else:
                return False
        except Exception as e:
            logger.warn(f"Failed to read 'class_to_id.json': {repr(e)}")
            return False

    def validate_format(self) -> bool:
        if self.upload_as_links:
            self._download_remote_ann_files()
        detected_ann_cnt = 0
        images_list, ann_dict = [], {}
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                dir_name = os.path.basename(root)
                file_name = get_file_name(full_path)
                if file.lower() == "class_to_id.json":
                    success = self.validate_key_file(os.path.join(root, file))
                    if not success:
                        logger.warn(
                            f"Failed to validate key file: '{file}'. Will use default cityscapes classes."
                        )
                if file in JUNK_FILES:
                    continue
                if file.endswith("_gtFine_polygons.json"):
                    success = self.validate_ann_file(full_path)
                    if success:
                        detected_ann_cnt += 1
                        ann_dict[file] = full_path
                if file_name.endswith("_leftImg8bit"):
                    if self.is_image(full_path):
                        images_list.append(full_path)

        meta = self._meta if self._meta is not None else ProjectMeta()
        self._items = []
        for image_path in images_list:
            image_name = get_file_name(image_path)
            if image_name.endswith("_leftImg8bit"):
                image_name = image_name[: -len("_leftImg8bit")]
            item = self.Item(image_path)
            ann_path = ann_dict.get(f"{image_name}_gtFine_polygons.json")
            if ann_path is not None:
                meta = self._generate_meta_from_ann(ann_path, meta)
                item.ann_data = ann_path
            self._items.append(item)

        if detected_ann_cnt > 0:
            tag_meta = TagMeta("split", TagValueType.ANY_STRING)
            if not meta.tag_metas.has_key(tag_meta.name):
                meta = meta.add_tag_meta(tag_meta)
            self._meta = meta
        return detected_ann_cnt > 0

    def _generate_meta_from_ann(self, ann_path: str, meta: ProjectMeta) -> ProjectMeta:
        ann_data = load_json_file(ann_path)
        for obj in ann_data["objects"]:
            class_name = obj["label"]
            obj_class = meta.get_obj_class(class_name)
            if obj_class is None:
                obj_class = ObjClass(class_name, Polygon)
                meta = meta.add_obj_class(obj_class)
        return meta

    def to_supervisely(
        self,
        item: ImageConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> Annotation:
        """Convert to Supervisely format."""
        if meta is None:
            meta = self._meta

        ann = item.create_empty_annotation()

        split_tag_name = renamed_tags.get("split", "split") if renamed_tags is not None else "split"
        tag_meta = meta.get_tag_meta(split_tag_name)
        path_parts = [ppart.lower() for ppart in Path(item.path).parts]
        if helper.VAL_TAG in path_parts:
            tag_value = helper.VAL_TAG
        elif helper.TEST_TAG in path_parts:
            tag_value = helper.TEST_TAG
        else:
            tag_value = helper.TRAIN_TAG
        split_tag = Tag(tag_meta, tag_value)

        ann = ann.add_tag(split_tag)

        if item.ann_data is None:
            return ann
        ann_path = item.ann_data
        try:
            if ann_path is not None:
                ann = helper.create_ann_from_file(ann, ann_path, meta, renamed_classes)
            return ann
        except Exception as e:
            logger.warn(f"Failed to convert annotation: {repr(e)}")
            return ann



# coding: utf-8

# docs
import base64
import errno
import hashlib
import mimetypes
import os
import re
import shutil
import subprocess
import tarfile
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Generator,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
)

import aiofiles
import requests
from requests.structures import CaseInsensitiveDict
from tqdm import tqdm

from supervisely._utils import get_bytes_hash, get_or_create_event_loop, get_string_hash

if TYPE_CHECKING:
    from supervisely.api.image_api import BlobImageInfo

from supervisely.io.fs_cache import FileCache
from supervisely.sly_logger import logger
from supervisely.task.progress import Progress

JUNK_FILES = [".DS_Store", "__MACOSX", "._.DS_Store", "Thumbs.db", "desktop.ini"]
OFFSETS_PKL_SUFFIX = "_offsets.pkl"  # suffix for pickle file with image offsets
OFFSETS_PKL_BATCH_SIZE = 10000  # 10k images per batch when loading from pickle


def get_file_name(path: str) -> str:
    """
    Extracts file name from a given path.

    :param path: Path to file.
    :type path: str
    :returns: File name without extension
    :rtype: :class:`str`
    :Usage example:

     .. code-block::

        import supervisely as sly

        file_name = sly.fs.get_file_name("/home/admin/work/projects/lemons_annotated/ds1/img/IMG_0748.jpeg")

        print(file_name)
        # Output: IMG_0748
    """
    return os.path.splitext(os.path.basename(path))[0]


def get_file_ext(path: str) -> str:
    """
    Extracts file extension from a given path.

    :param path: Path to file.
    :type path: str
    :returns: File extension without name
    :rtype: :class:`str`
    :Usage example:

     .. code-block::

        import supervisely as sly

        file_ext = sly.fs.get_file_ext("/home/admin/work/projects/lemons_annotated/ds1/img/IMG_0748.jpeg")

        print(file_ext)
        # Output: .jpeg
    """
    return os.path.splitext(os.path.basename(path))[1]


def get_file_name_with_ext(path: str) -> str:
    """
    Extracts file name with ext from a given path.

    :param path: Path to file.
    :type path: str
    :returns: File name with extension
    :rtype: :class:`str`
    :Usage example:

     .. code-block::

        import supervisely as sly

        file_name_ext = sly.fs.get_file_name_with_ext("/home/admin/work/projects/lemons_annotated/ds1/img/IMG_0748.jpeg")

        print(file_name_ext)
        # Output: IMG_0748.jpeg
    """
    return os.path.basename(path)


def remove_junk_from_dir(dir: str) -> List[str]:
    """
    Cleans the given directory from junk files and dirs (e.g. .DS_Store, __MACOSX, Thumbs.db, etc.).

    :param dir: Path to directory.
    :type dir: str
    :returns: List of global paths to removed files and dirs.
    :rtype: List[str]

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        input_dir = "/home/admin/work/projects/lemons_annotated/"
        sly.fs.remove_junk_from_dir(input_dir)
    """
    paths = list_dir_recursively(dir, include_subdirs=True, use_global_paths=True)
    removed_paths = []
    for path in paths:
        if get_file_name(path) in JUNK_FILES:
            if os.path.isfile(path):
                silent_remove(path)
                removed_paths.append(path)
            elif os.path.isdir(path):
                remove_dir(path)
                removed_paths.append(path)


def list_dir_recursively(
    dir: str, include_subdirs: bool = False, use_global_paths: bool = False
) -> List[str]:
    """
    Recursively walks through directory and returns list with all file paths, and optionally subdirectory paths.

    :param path: Path to directory.
    :type path: str
    :param include_subdirs: If True, subdirectory paths will be included in the result list.
    :type include_subdirs: bool
    :param use_global_paths: If True, absolute paths will be returned instead of relative ones.
    :type use_global_paths: bool
    :returns: List containing file paths, and optionally subdirectory paths.
    :rtype: :class:`List[str]`
    :Usage example:

     .. code-block::

        import supervisely as sly

        list_dir = sly.fs.list_dir_recursively("/home/admin/work/projects/lemons_annotated/")

        print(list_dir)
        # Output: ['meta.json', 'ds1/ann/IMG_0748.jpeg.json', 'ds1/ann/IMG_4451.jpeg.json', 'ds1/img/IMG_0748.jpeg', 'ds1/img/IMG_4451.jpeg']
    """
    all_paths = []
    for root, dirs, files in os.walk(dir):
        for name in files:
            file_path = os.path.join(root, name)
            file_path = (
                os.path.relpath(file_path, dir)
                if not use_global_paths
                else os.path.abspath(file_path)
            )
            all_paths.append(file_path)
        if include_subdirs:
            for name in dirs:
                subdir_path = os.path.join(root, name)
                subdir_path = (
                    os.path.relpath(subdir_path, dir)
                    if not use_global_paths
                    else os.path.abspath(subdir_path)
                )
                all_paths.append(subdir_path)
    return all_paths


def list_files_recursively(
    dir: str,
    valid_extensions: Optional[List[str]] = None,
    filter_fn=None,
    ignore_valid_extensions_case: Optional[bool] = False,
) -> List[str]:
    """
    Recursively walks through directory and returns list with all file paths.
    Can be filtered by valid extensions and filter function.

    :param dir: Target dir path.
    :param dir: str
    :param valid_extensions: List with valid file extensions.
    :type valid_extensions: List[str], optional
    :param filter_fn: Function with a single argument. Argument is a file path. Function determines whether to keep a given file path. Must return True or False.
    :type filter_fn: Callable, optional
    :param ignore_valid_extensions_case: If True, validation of file extensions will be case insensitive.
    :type ignore_valid_extensions_case: bool
    :returns: List with file paths
    :rtype: :class:`List[str]`
    :Usage example:

     .. code-block:: python

         import supervisely as sly

         list_files = sly.fs.list_files_recursively("/home/admin/work/projects/lemons_annotated/ds1/img/")

         print(list_files)
         # Output: ['/home/admin/work/projects/lemons_annotated/ds1/img/IMG_0748.jpeg', '/home/admin/work/projects/lemons_annotated/ds1/img/IMG_4451.jpeg']
    """

    def file_path_generator():
        for dir_name, _, file_names in os.walk(dir):
            for filename in file_names:
                yield os.path.join(dir_name, filename)

    valid_extensions = (
        valid_extensions
        if ignore_valid_extensions_case is False
        else [ext.lower() for ext in valid_extensions]
    )
    files = []
    for file_path in file_path_generator():
        file_ext = get_file_ext(file_path)
        if ignore_valid_extensions_case:
            file_ext.lower()
        if (valid_extensions is None or file_ext in valid_extensions) and (
            filter_fn is None or filter_fn(file_path)
        ):
            files.append(file_path)
    return files


def list_files(
    dir: str,
    valid_extensions: Optional[List[str]] = None,
    filter_fn=None,
    ignore_valid_extensions_case: Optional[bool] = False,
) -> List[str]:
    """
    Returns list with file paths presented in given directory.
    Can be filtered by valid extensions and filter function.
    Also can be case insensitive for valid extensions.

    :param dir: Target dir path.
    :param dir: str
    :param valid_extensions: List with valid file extensions.
    :type valid_extensions: List[str]
    :param filter_fn: Function with a single argument. Argument is a file path. Function determines whether to keep a given file path. Must return True or False.
    :type filter_fn: Callable, optional
    :param ignore_valid_extensions_case: If True, validation of file extensions will be case insensitive.
    :type ignore_valid_extensions_case: bool
    :returns: List with file paths
    :rtype: :class:`List[str]`
    :Usage example:

     .. code-block:: python

         import supervisely as sly

         list_files = sly.fs.list_files("/home/admin/work/projects/lemons_annotated/ds1/img/")

         print(list_files)
         # Output: ['/home/admin/work/projects/lemons_annotated/ds1/img/IMG_0748.jpeg', '/home/admin/work/projects/lemons_annotated/ds1/img/IMG_4451.jpeg']
    """
    res = list(os.path.join(dir, x.name) for x in os.scandir(dir) if x.is_file())

    files = []
    for file_path in res:
        file_ext = get_file_ext(file_path)

        if ignore_valid_extensions_case:
            file_ext = file_ext.lower()
            valid_extensions = [ext.lower() for ext in valid_extensions]

        if (valid_extensions is None or file_ext in valid_extensions) and (
            filter_fn is None or filter_fn(file_path)
        ):
            files.append(file_path)

    return files

    # return [
    #     file_path
    #     for file_path in res
    #     if (valid_extensions is None or get_file_ext(file_path) in valid_extensions)
    #     and (filter_fn is None or filter_fn(file_path))
    # ]


def mkdir(dir: str, remove_content_if_exists: Optional[bool] = False) -> None:
    """
    Creates a leaf directory and all intermediate ones.

    :param dir: Target dir path.
    :param dir: str
    :remove_content_if_exists: Remove directory content if it exist.
    :remove_content_if_exists: bool
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import mkdir
        mkdir('/home/admin/work/projects/example')
    """
    if dir_exists(dir) and remove_content_if_exists is True:
        clean_dir(dir, ignore_errors=True)
    else:
        os.makedirs(dir, mode=0o777, exist_ok=True)


# create base path recursively; pay attention to slash-terminating paths
def ensure_base_path(path: str) -> None:
    """
    Recursively create parent directory for target path.

    :param path: Target dir path.
    :type path: str
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import ensure_base_path
        ensure_base_path('/home/admin/work/projects/example')
    """
    dst_dir = os.path.split(path)[0]
    if dst_dir:
        mkdir(dst_dir)


def copy_file(src: str, dst: str) -> None:
    """
    Copy file from one path to another, if destination directory doesn't exist it will be created.

    :param src: Source file path.
    :type src: str
    :param dst: Destination file path.
    :type dst: str
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import copy_file
        copy_file('/home/admin/work/projects/example/1.png', '/home/admin/work/tests/2.png')
    """
    ensure_base_path(dst)
    with open(dst, "wb") as out_f:
        with open(src, "rb") as in_f:
            shutil.copyfileobj(in_f, out_f, length=1024 * 1024)


def hardlink_or_copy_file(src: str, dst: str) -> None:
    """
    Creates a hard link pointing to src named dst. If the link cannot be created, the file will be copied.

    :param src: Source file path.
    :type src: str
    :param dst: Destination file path.
    :type dst: str
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import hardlink_or_copy_file
        hardlink_or_copy_file('/home/admin/work/projects/example/1.png', '/home/admin/work/tests/link.txt')
    """
    try:
        os.link(src, dst)
    except OSError:
        copy_file(src, dst)


def hardlink_or_copy_tree(src: str, dst: str) -> None:
    """
    Creates a hard links pointing to src named dst files recursively. If the link cannot be created, the file will be copied.

    :param src: Source dir path.
    :type src: str
    :param dst: Destination dir path.
    :type dst: str
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import hardlink_or_copy_tree
        hardlink_or_copy_tree('/home/admin/work/projects/examples', '/home/admin/work/tests/links')
    """
    mkdir(dst)
    for dir_name, _, file_names in os.walk(src):
        relative_dir = os.path.relpath(dir_name, src)
        dst_sub_dir = os.path.join(dst, relative_dir)
        mkdir(dst_sub_dir)
        for file_name in file_names:
            hardlink_or_copy_file(
                os.path.join(dir_name, file_name), os.path.join(dst_sub_dir, file_name)
            )


def dir_exists(dir: str) -> bool:
    """
    Check whether directory exists or not.

    :param dir: Target directory path.
    :type dir: str
    :returns: True if directory exists, False otherwise.
    :rtype: :class:`bool`
    :Usage example:

     .. code-block:: python

          from supervisely.io.fs import dir_exists
          dir_exists('/home/admin/work/projects/examples') # True
          dir_exists('/home/admin/work/not_exist_dir') # False
    """
    return os.path.isdir(dir)


def dir_empty(dir: str) -> bool:
    """
    Check whether directory is empty or not.

    :param dir: Target directory path.
    :type dir: str
    :returns: True if directory is empty, False otherwise.
    :rtype: :class:`bool`
    :Usage example:

     .. code-block:: python

          from supervisely.io.fs import dir_empty
          dir_empty('/home/admin/work/projects/examples') # False
    """
    if dir_exists(dir) and len(list_files_recursively(dir)) > 0:
        return False
    return True


def file_exists(path: str) -> bool:
    """
    Check whether file exists or not.

    :param dir: Target file path.
    :type dir: str
    :returns: True if file exists, False otherwise.
    :rtype: :class:`bool`
    :Usage example:

     .. code-block:: python

          from supervisely.io.fs import file_exists
          file_exists('/home/admin/work/projects/examples/1.jpeg') # True
          file_exists('/home/admin/work/projects/examples/not_exist_file.jpeg') # False
    """
    return os.path.isfile(path)


def get_subdirs(dir_path: str, recursive: Optional[bool] = False) -> list:
    """
    Get list containing the names of the directories in the given directory.

    :param dir_path: Target directory path.
    :type dir_path: str
    :param recursive: If True, all found subdirectories will be included in the result list.
    :type recursive: bool
    :returns: List containing directories names.
    :rtype: :class:`list`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import get_subdirs
        subdirs = get_subdirs('/home/admin/work/projects/examples')
        print(subdirs)
        # Output: ['tests', 'users', 'ds1']
    """
    if recursive:
        return [
            global_to_relative(entry, dir_path)
            for entry in list_dir_recursively(dir_path, include_subdirs=True, use_global_paths=True)
            if os.path.isdir(entry)
        ]
    res = list(x.name for x in os.scandir(dir_path) if x.is_dir())
    return res


def get_subdirs_tree(dir_path: str) -> Dict[str, Union[str, Dict]]:
    """Returns a dictionary representing the directory tree.
    It will have only directories and subdirectories (not files).

    :param dir_path: Target directory path.
    :type dir_path: str
    :returns: Dictionary representing the directory tree.
    :rtype: :class:`Dict[str, Union[str, Dict]]`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import get_subdirs_tree
        tree = get_subdirs_tree('/home/admin/work/projects/examples')
        print(tree)
        # Output: {'examples': {'tests': {}, 'users': {}, 'ds1': {}}}
    """

    tree = {}
    subdirs = get_subdirs(dir_path, recursive=True)
    for subdir in subdirs:
        parts = subdir.split(os.sep)
        d = tree
        for part in parts:
            if part not in d:
                d[part] = {}
            d = d[part]

    return tree


def subdirs_tree(
    dir_path: str,
    ignore: Optional[List[str]] = None,
    ignore_content: Optional[List[str]] = None,
) -> Generator[str, None, None]:
    """Generator that yields directories in the directory tree,
    starting from the level below the root directory and then going down the tree.
    If ignore is specified, it will ignore paths which end with the specified directory names.
    All subdirectories of ignored directories will still be yielded.

    :param dir_path: Target directory path.
    :type dir_path: str
    :param ignore: List of directories to ignore. Note, that function still will yield
        subdirectories of ignored directories. It will only ignore paths which end with
        the specified directory names.
    :type ignore: List[str]
    :param ignore_content: List of directories which subdirectories should be ignored.
    :type ignore_content: List[str]
    :returns: Generator that yields directories in the directory tree.
    :rtype: Generator[str, None, None]
    """
    tree = get_subdirs_tree(dir_path)
    ignore = ignore or []
    ignore_content = ignore_content or []

    def _subdirs_tree(tree, path=""):
        for key, value in tree.items():
            new_path = os.path.join(path, key) if path else key
            if not any(new_path.endswith(i) for i in ignore):
                yield new_path
            if any(new_path.endswith(i) for i in ignore_content):
                continue
            if value:
                yield from _subdirs_tree(value, new_path)

    yield from _subdirs_tree(tree)


def global_to_relative(global_path: str, base_dir: str) -> str:
    """
    Converts global path to relative path.

    :param global_path: Global path.
    :type global_path: str
    :param base_dir: Base directory path.
    :type base_dir: str
    :returns: Relative path.
    :rtype: :class:`str`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import global_to_relative
        relative_path = global_to_relative('/home/admin/work/projects/examples/1.jpeg', '/home/admin/work/projects')
        print(relative_path)
        # Output: examples/1.jpeg
    """
    return os.path.relpath(global_path, base_dir)


# removes directory content recursively
def clean_dir(dir_: str, ignore_errors: Optional[bool] = True) -> None:
    """
    Recursively delete a directory tree, but save root directory.

    :param dir_: Target directory path.
    :type dir_: str
    :ignore_errors: Ignore possible errors while removes directory content.
    :ignore_errors: bool
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import clean_dir
        clean_dir('/home/admin/work/projects/examples')
    """
    # old implementation
    # shutil.rmtree(dir_, ignore_errors=True)
    # mkdir(dir_)

    for filename in os.listdir(dir_):
        file_path = os.path.join(dir_, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            logger.warn(f"Failed to delete {file_path}. Reason: {repr(e)}")
            if ignore_errors is False:
                raise e


def remove_dir(dir_: str) -> None:
    """
    Recursively delete a directory tree.

    :param dir_: Target directory path.
    :type dir_: str
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import remove_dir
        remove_dir('/home/admin/work/projects/examples')
    """
    shutil.rmtree(dir_, ignore_errors=True)


def silent_remove(file_path: str) -> None:
    """
    Remove file which may not exist.

    :param file_path: File path.
    :type file_path: str
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import silent_remove
        silent_remove('/home/admin/work/projects/examples/1.jpeg')
    """
    try:
        os.remove(file_path)
    except OSError as e:
        if e.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
            raise


def get_file_size(path: str) -> int:
    """
    Get the size of a file.

    :param path: File path.
    :type path: str
    :returns: File size in bytes
    :rtype: :class:`int`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import get_file_size
        file_size = get_file_size('/home/admin/work/projects/examples/1.jpeg') # 161665
    """
    return os.path.getsize(path)


def get_directory_size(dir_path: str) -> int:
    """
    Get the size of a directory.

    :param path: Target directory path.
    :type path: str
    :returns: Directory size in bytes
    :rtype: :class:`int`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import get_directory_size
        dir_size = get_directory_size('/home/admin/work/projects/examples') # 8574563
    """
    total_size = 0
    for dirpath, _, filenames in os.walk(dir_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += get_file_size(fp)
    return total_size


def archive_directory(
    dir_: str, tar_path: str, split: Optional[Union[int, str]] = None, chunk_size_mb: int = 50
) -> Union[None, List[str]]:
    """
    Create tar archive from directory and optionally split it into parts of specified size.
    You can adjust the size of the chunk to read from the file, while archiving the file into parts.
    Be careful with this parameter, it can affect the performance of the function.
    When spliting, if the size of split is less than the chunk size, the chunk size will be adjusted to fit the split size.

    :param dir_: Target directory path.
    :type dir_: str
    :param tar_path: Path for output tar archive.
    :type tar_path: str
    :param split: Split archive into parts of specified size (in bytes) or size with
        suffix (e.g. '1Kb' = 1024, '1Mb' = 1024 * 1024). Default is None.
    :type split: Union[int, str]
    :param chunk_size_mb: Size of the chunk to read from the file. Default is 50Mb.
    :type chunk_size_mb: int
    :returns: None or list of archive parts if split is not None
    :rtype: Union[None, List[str]]
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import archive_directory
        # If split is not needed.
        archive_directory('/home/admin/work/projects/examples', '/home/admin/work/examples.tar')

        # If split is specified.
        archive_parts_paths = archive_directory('/home/admin/work/projects/examples', '/home/admin/work/examples/archive.tar', split=1000000)
        print(archive_parts_paths) # ['/home/admin/work/examples/archive.tar.001', '/home/admin/work/examples/archive.tar.002']
    """
    with tarfile.open(tar_path, "w", encoding="utf-8") as tar:
        tar.add(dir_, arcname=os.path.sep)

    if split is None:
        return

    split = string_to_byte_size(split)

    if os.path.getsize(tar_path) <= split:
        return

    chunk = chunk_size_mb * 1024 * 1024
    tar_name = os.path.basename(tar_path)
    tar_dir = os.path.abspath(os.path.dirname(tar_path))
    parts_paths = []
    part_number = 1

    original_chunk = chunk
    chunk = min(chunk, split)  # chunk size should be less than split size
    if chunk != original_chunk:
        logger.info(f"Chunk size adjusted to {chunk} bytes to fit split size.")

    with open(tar_path, "rb") as input_file:
        while True:
            part_name = f"{tar_name}.{str(part_number).zfill(3)}"
            output_path = os.path.join(tar_dir, part_name)
            with open(output_path, "wb") as output_file:
                while output_file.tell() < split:
                    data = input_file.read(chunk)
                    if not data:
                        break
                    output_file.write(data)
                if not data and output_file.tell() == 0:
                    os.remove(output_path)
                    break
                parts_paths.append(output_path)
                part_number += 1
            if not data:
                break

    os.remove(tar_path)
    return parts_paths


def unpack_archive(
    archive_path: str, target_dir: str, remove_junk=True, is_split=False, chunk_size_mb: int = 50
) -> None:
    """
    Unpacks archive to the target directory, removes junk files and directories.
    To extract a split archive, you must pass the path to the first part in archive_path. Archive parts must be in the same directory. Format: archive_name.tar.001, archive_name.tar.002, etc. Works with tar and zip.
    You can adjust the size of the chunk to read from the file, while unpacking the file from parts.
    Be careful with this parameter, it can affect the performance of the function.

    :param archive_path: Path to the archive.
    :type archive_path: str
    :param target_dir: Path to the target directory.
    :type target_dir: str
    :param remove_junk: Remove junk files and directories. Default is True.
    :type remove_junk: bool
    :param is_split: Determines if the source archive is split into parts. If True, archive_path must be the path to the first part. Default is False.
    :type is_split: bool
    :param chunk_size_mb: Size of the chunk to read from the file. Default is 50Mb.
    :type chunk_size_mb: int
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        archive_path = '/home/admin/work/examples.tar'
        target_dir = '/home/admin/work/projects'

        sly.fs.unpack_archive(archive_path, target_dir)
    """

    if is_split:
        chunk = chunk_size_mb * 1024 * 1024
        base_name = get_file_name(archive_path)
        dir_name = os.path.dirname(archive_path)
        if get_file_ext(base_name) in (".zip", ".tar"):
            ext = get_file_ext(base_name)
            base_name = get_file_name(base_name)
        else:
            ext = get_file_ext(archive_path)
        parts = sorted([f for f in os.listdir(dir_name) if f.startswith(base_name)])
        combined = os.path.join(dir_name, f"combined{ext}")

        with open(combined, "wb") as output_file:
            for part in parts:
                part_path = os.path.join(dir_name, part)
                with open(part_path, "rb") as input_file:
                    while True:
                        data = input_file.read(chunk)
                        if not data:
                            break
                        output_file.write(data)
        archive_path = combined

    shutil.unpack_archive(archive_path, target_dir)
    if is_split:
        silent_remove(archive_path)
    if remove_junk:
        remove_junk_from_dir(target_dir)


def string_to_byte_size(string: Union[str, int]) -> int:
    """Returns integer representation of byte size from string representation.
        If input is integer, returns the same integer for convenience.

        :param string: string representation of byte size (e.g. 1.5Kb, 2Mb, 3.7Gb, 4.2Tb) or integer
        :type string: Union[str, int]
        :return: integer representation of byte size (or the same integer if input is integer)
        :rtype: int

        :raises ValueError: if input string is invalid

    :Usage example:

    .. code-block:: python
        string_size = "1.5M"
        size = string_to_byte_size(string_size)
        print(size)  # 1572864

    """

    MULTIPLIER = 1024
    units = {"KB": 1, "MB": 2, "GB": 3, "TB": 4}

    if isinstance(string, int):
        return string

    try:
        value, unit = string[:-2], string[-2:].upper()
        multiplier = MULTIPLIER ** units[unit]
        return int(float(value) * multiplier)
    except (KeyError, ValueError, IndexError):
        raise ValueError(
            "Invalid input string. The string must be in the format of '1.5Kb', '2Mb', '3.7Gb', '4.2Tb' or integer."
        )


def get_file_hash(path: str) -> str:
    """
    Get hash from target file.

    :param path: Target file path.
    :type path: str
    :returns: File hash
    :rtype: :class:`str`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import get_file_hash
        hash = get_file_hash('/home/admin/work/projects/examples/1.jpeg') # rKLYA/p/P64dzidaQ/G7itxIz3ZCVnyUhEE9fSMGxU4=
    """
    with open(path, "rb") as file:
        file_bytes = file.read()
        return get_bytes_hash(file_bytes)


def get_file_hash_chunked(path: str, chunk_size: Optional[int] = 1024 * 1024) -> str:
    """
    Get hash from target file by reading it in chunks.

    :param path: Target file path.
    :type path: str
    :param chunk_size: Number of bytes to read per iteration. Default is 1 MB.
    :type chunk_size: int, optional
    :returns: File hash as a base64 encoded string.
    :rtype: str

    :Usage example:

    .. code-block:: python

       file_hash = sly.fs.get_file_hash_chunked('/home/admin/work/projects/examples/1.jpeg')
       print(file_hash)  # Example output: rKLYA/p/P64dzidaQ/G7itxIz3ZCVnyUhEE9fSMGxU4=
    """
    hash_sha256 = hashlib.sha256()
    with open(path, "rb") as f:
        while chunk := f.read(chunk_size):
            hash_sha256.update(chunk)
    digest = hash_sha256.digest()
    return base64.b64encode(digest).decode("utf-8")


def tree(dir_path: str) -> str:
    """
    Get tree for target directory.

    :param dir_path: Target directory path.
    :type dir_path: str
    :returns: Tree with directory files and subdirectories
    :rtype: :class:`str`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import tree
        dir_tree = tree('/home/admin/work/projects/examples')
        print(dir_tree)
        # Output: /home/admin/work/projects/examples
        # ├── [4.0K]  1
        # │   ├── [165K]  crop.jpeg
        # │   ├── [169K]  fliplr.jpeg
        # │   ├── [169K]  flipud.jpeg
        # │   ├── [166K]  relative_crop.jpeg
        # │   ├── [167K]  resize.jpeg
        # │   ├── [169K]  rotate.jpeg
        # │   ├── [171K]  scale.jpeg
        # │   └── [168K]  translate.jpeg
        # ├── [ 15K]  123.jpeg
        # ├── [158K]  1.jpeg
        # ├── [188K]  1.txt
        # ├── [1.3M]  1.zip
        # ├── [4.0K]  2
        # ├── [ 92K]  acura.png
        # ├── [1.2M]  acura_PNG122.png
        # ├── [198K]  aston_martin_PNG55.png
        # ├── [4.0K]  ds1
        # │   ├── [4.0K]  ann
        # │   │   ├── [4.3K]  IMG_0748.jpeg.json
        # │   │   ├── [ 151]  IMG_0777.jpeg.json
        # │   │   ├── [ 151]  IMG_0888.jpeg.json
        # │   │   ├── [3.7K]  IMG_1836.jpeg.json
        # │   │   ├── [8.1K]  IMG_2084.jpeg.json
        # │   │   ├── [5.5K]  IMG_3861.jpeg.json
        # │   │   ├── [6.0K]  IMG_4451.jpeg.json
        # │   │   └── [5.0K]  IMG_8144.jpeg.json
        # │   └── [4.0K]  img
        # │       ├── [152K]  IMG_0748.jpeg
        # │       ├── [210K]  IMG_0777.jpeg
        # │       ├── [210K]  IMG_0888.jpeg
        # │       ├── [137K]  IMG_1836.jpeg
        # │       ├── [139K]  IMG_2084.jpeg
        # │       ├── [145K]  IMG_3861.jpeg
        # │       ├── [133K]  IMG_4451.jpeg
        # │       └── [136K]  IMG_8144.jpeg
        # ├── [152K]  example.jpeg
        # ├── [2.4K]  example.json
        # ├── [153K]  flip.jpeg
        # ├── [ 65K]  hash1.jpeg
        # ├── [ 336]  meta.json
        # └── [5.4K]  q.jpeg
        # 5 directories, 37 files
    """
    out = subprocess.Popen(
        ["tree", "--filelimit", "500", "-h", "-n", dir_path],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    stdout, stderr = out.communicate()
    return stdout.decode("utf-8")


def log_tree(
    dir_path: str,
    logger,
    level: Literal["info", "debug", "warning", "error"] = "info",
) -> None:
    """
    Get tree for target directory and displays it in the log.

    :param dir_path: Target directory path.
    :type dir_path: str
    :param logger: Logger to display data.
    :type logger: logger
    :type level: Logger level. Available levels: info, debug, warning, error. Default: info.
    :type level: Literal["info", "debug", "warning", "error"]
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import log_tree
        logger = sly.logger
        log_tree('/home/admin/work/projects/examples', logger)
    """
    out = tree(dir_path)

    log_levels = {
        "info": logger.info,
        "debug": logger.debug,
        "warning": logger.warning,
        "error": logger.error,
    }
    if level not in log_levels:
        raise ValueError(
            f"Unknown logger level: {level}. Available levels: info, debug, warning, error"
        )
    log_func = log_levels[level]
    log_func("DIRECTORY_TREE", extra={"tree": out})


def touch(path: str) -> None:
    """
    Sets access and modification times for a file.

    :param path: Target file path.
    :type path: str
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import touch
        touch('/home/admin/work/projects/examples/1.jpeg')
    """
    ensure_base_path(path)
    with open(path, "a"):
        os.utime(path, None)


def download(
    url: str,
    save_path: str,
    cache: Optional[FileCache] = None,
    progress: Optional[Callable] = None,
    headers: Optional[Dict] = None,
    timeout: Optional[int] = None,
) -> str:
    """
    Load image from url to host by target path.

    :param url: Target file path.
    :type url: str
    :param url: The path where the file is saved.
    :type url: str
    :param cache: An instance of FileCache class that provides caching functionality for the downloaded content. If None, caching is disabled.
    :type cache: FileCache, optional
    :param progress: Function for tracking download progress.
    :type progress: Progress, optional
    :param headers: A dictionary of HTTP headers to include in the request.
    :type headers: Dict, optional.
    :param timeout: The maximum number of seconds to wait for a response from the server. If the server does not respond within the timeout period, a TimeoutError is raised.
    :type timeout: int, optional.
    :returns: Full path to downloaded image
    :rtype: :class:`str`
    :Usage example:

     .. code-block:: python

        from supervisely.io.fs import download
        img_link = 'https://m.media-amazon.com/images/M/MV5BMTYwOTEwNjAzMl5BMl5BanBnXkFtZTcwODc5MTUwMw@@._V1_.jpg'
        im_path = download(img_link, '/home/admin/work/projects/examples/avatar.jpeg')
        print(im_path)
        # Output:
        # /home/admin/work/projects/examples/avatar.jpeg

        # if you need to specify some headers
        headers = {
            'User-Agent': 'Mozilla/5.0',
        }
        im_path = download(img_link, '/home/admin/work/projects/examples/avatar.jpeg', headers=headers)
        print(im_path)
        # Output:
        # /home/admin/work/projects/examples/avatar.jpeg

    """

    def _download():
        try:
            with requests.get(url, stream=True, headers=headers, timeout=timeout) as r:
                r.raise_for_status()
                total_size_in_bytes = int(CaseInsensitiveDict(r.headers).get("Content-Length", "0"))
                if progress is not None and type(progress) is Progress:
                    progress.set(0, total_size_in_bytes)
                with open(save_path, "wb") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)
                        if progress is not None:
                            if type(progress) is Progress:
                                progress.iters_done_report(len(chunk))
                            else:
                                progress(len(chunk))
        except requests.exceptions.Timeout as e:
            message = (
                "Request timed out. "
                "This may be due to server-side security measures, network congestion, or other issues. "
                "Please check your server logs for more information."
            )
            logger.warn(msg=message)
            raise e

    if cache is None:
        _download()
    else:
        cache_path = cache.check_storage_object(get_string_hash(url), get_file_ext(save_path))
        if cache_path is None:
            # file not in cache
            _download()
            cache.write_object(save_path, get_string_hash(url))
        else:
            cache.read_object(get_string_hash(url), save_path)
            if progress is not None:
                sizeb = get_file_size(save_path)
                if type(progress) is Progress:
                    progress.set(0, sizeb)
                    progress.iters_done_report(sizeb)
                else:
                    progress(sizeb)

    return save_path


def copy_dir_recursively(
    src_dir: str, dst_dir: str, progress_cb: Optional[Union[tqdm, Callable]] = None
) -> List[str]:
    mkdir(dst_dir)
    src_dir_norm = src_dir.rstrip(os.sep)

    for rel_sub_dir in get_subdirs(src_dir_norm, recursive=True):
        dst_sub_dir = os.path.join(dst_dir, rel_sub_dir)
        mkdir(dst_sub_dir)

    files = list_files_recursively(src_dir_norm)
    for src_file_path in files:
        dst_file_path = os.path.normpath(src_file_path.replace(src_dir_norm, dst_dir))
        ensure_base_path(dst_file_path)
        if not file_exists(dst_file_path):
            copy_file(src_file_path, dst_file_path)
            if progress_cb is not None:
                progress_cb(get_file_size(src_file_path))


def is_on_agent(remote_path: str) -> bool:
    """Check if remote_path starts is on agent (e.g. starts with 'agent://<agent-id>/').

    :param remote_path: path to check
    :type remote_path: str
    :return: True if remote_path starts with 'agent://<agent-id>/' and False otherwise
    :rtype: bool
    """
    if remote_path.startswith("agent://"):
        return True
    else:
        return False


def parse_agent_id_and_path(remote_path: str) -> Tuple[int, str]:
    """Return agent id and path in agent folder from remote_path.

    :param remote_path: path to parse
    :type remote_path: str
    :return: agent id and path in agent folder
    :rtype: Tuple[int, str]
    :raises ValueError: if remote_path doesn't start with 'agent://<agent-id>/'
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Parse agent id and path in agent folder from remote_path
        remote_path = "agent://1/agent_folder/subfolder/file.txt"
        agent_id, path_in_agent_folder = sly.fs.parse_agent_id_and_path(remote_path)
        print(agent_id)  # 1
        print(path_in_agent_folder)  # /agent_folder/subfolder/file.txt
    """
    if is_on_agent(remote_path) is False:
        raise ValueError("agent path have to starts with 'agent://<agent-id>/'")
    search = re.search("agent://(\d+)(.*)", remote_path)
    agent_id = int(search.group(1))
    path_in_agent_folder = search.group(2)
    if not path_in_agent_folder.startswith("/"):
        path_in_agent_folder = "/" + path_in_agent_folder
    if remote_path.endswith("/") and not path_in_agent_folder.endswith("/"):
        path_in_agent_folder += "/"
    # path_in_agent_folder = os.path.normpath(path_in_agent_folder)
    return agent_id, path_in_agent_folder


def dirs_with_marker(
    input_path: str,
    markers: Union[str, List[str]],
    check_function: Callable = None,
    ignore_case: bool = False,
) -> Generator[str, None, None]:
    """
    Generator that yields paths to directories that contain markers files. If the check_function is specified,
    then the markered directory will be yielded only if the check_function returns True. The check_function
    must take a single argument - the path to the markered directory and return True or False.

    :param input_path: path to the directory in which the search will be performed
    :type input_path: str
    :param markers: single marker or list of markers (e.g. 'config.json' or ['config.json', 'config.yaml'])
    :type markers: Union[str, List[str]]
    :param check_function: function to check that directory meets the requirements and returns bool
    :type check_function: Callable
    :param ignore_case: ignore case when searching for markers
    :type ignore_case: bool
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        input_path = '/home/admin/work/projects/examples'

        # You can pass a string if you have only one marker.
        # markers = 'config.json'

        # Or a list of strings if you have several markers.
        # There's no need to pass one marker in different cases, you can use ignore_case=True for this.
        markers = ['config.json', 'config.yaml']


        # Check function is optional, if you don't need the directories to meet any requirements,
        # you can omit it.

        def check_function(dir_path):
            test_file_path = os.path.join(dir_path, 'test.txt')
            return os.path.exists(test_file_path)

        for directory in sly.fs.dirs_with_marker(input_path, markers, check_function, ignore_case=True):
            # Now you can be sure that the directory contains the markers and meets the requirements.
            # Do something with it.
            print(directory)
    """

    if isinstance(markers, str):
        markers = [markers]

    paths = list_dir_recursively(input_path)
    for path in paths:
        for marker in markers:
            filename = get_file_name_with_ext(path)
            if ignore_case:
                filename = filename.lower()
                marker = marker.lower()

            if filename == marker:
                parent_dir = os.path.dirname(path)
                markered_dir = os.path.join(input_path, parent_dir)

                if check_function is None or check_function(markered_dir):
                    yield markered_dir


def dirs_filter(input_path: str, check_function: Callable) -> Generator[str, None, None]:
    """
    Generator that yields paths to directories that meet the requirements of the check_function.

    :param input_path: path to the directory in which the search will be performed
    :type input_path: str
    :param check_function: function to check that directory meets the requirements and returns bool
    :type check_function: Callable
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        input_path = '/home/admin/work/projects/examples'

        # Prepare the check function.

        def check_function(directory) -> bool:
            images_dir = os.path.join(directory, "images")
            annotations_dir = os.path.join(directory, "annotations")

            return os.path.isdir(images_dir) and os.path.isdir(annotations_dir)

        for directory in sly.fs.dirs(input_path, check_function):
            # Now you can be sure that the directory meets the requirements.
            # Do something with it.
            print(directory)
    """
    paths = [os.path.abspath(input_path)]
    paths.extend(list_dir_recursively(input_path, include_subdirs=True, use_global_paths=True))
    for path in paths:
        if os.path.isdir(path):
            if check_function(path):
                yield path


def change_directory_at_index(path: str, dir_name: str, dir_index: int) -> str:
    """
    Change directory name in path by index.
    If you use counting from the end, keep in mind that if the path ends with a file, the file will be assigned to the last index.

    :param path: The original path
    :type path: str
    :param dir_name: Directory name
    :type dir_name: str
    :param dir_index: Index of the directory we want to change, negative values count from the end
    :type dir_index: int
    :return: New path
    :rtype: str
    :raises IndexError: If the catalog index is out of bounds for a given path
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        input_path = 'head/dir_1/file.txt'
        new_path = sly.io.fs.change_directory_at_index(input_path, 'dir_2', -2)

        print(new_path)

    """
    path_components = path.split(os.path.sep)
    if -len(path_components) <= dir_index < len(path_components):
        path_components[dir_index] = dir_name
    else:
        raise IndexError(
            f"Path index '{dir_index}' is out of bounds 'path_components={len(path_components)}' for a given path"
        )
    return os.path.sep.join(path_components)


def is_archive(file_path: str) -> bool:
    """
    Checks if the file is an archive by its mimetype using list of the most common archive mimetypes.

    :param local_path: path to the local file
    :type local_path: str
    :return: True if the file is an archive, False otherwise
    :rtype: bool
    """
    archive_mimetypes = [
        "application/zip",
        "application/x-tar",
        "application/x-gzip",
        "application/x-bzip2",
        "application/x-7z-compressed",
        "application/x-rar-compressed",
        "application/x-xz",
        "application/x-lzip",
        "application/x-lzma",
        "application/x-lzop",
        "application/x-bzip",
        "application/x-bzip2",
        "application/x-compress",
        "application/x-compressed",
    ]

    return mimetypes.guess_type(file_path)[0] in archive_mimetypes


def str_is_url(string: str) -> bool:
    """
    Check if string is a valid URL.
    :param string: string to check
    :type string: str
    :return: True if string is a valid URL, False otherwise
    :rtype: bool
    :Usage example:
     .. code-block:: python
        import supervisely as sly
        url = 'https://example.com/image.jpg'
        is_url = sly.fs.str_is_url(url)
        print(is_url)  # True
    """
    from urllib.parse import urlparse

    try:
        result = urlparse(string)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


async def copy_file_async(
    src: str,
    dst: str,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
    progress_cb_type: Literal["number", "size"] = "size",
) -> None:
    """
    Asynchronously copy file from one path to another, if destination directory doesn't exist it will be created.

    :param src: Source file path.
    :type src: str
    :param dst: Destination file path.
    :type dst: str
    :param progress_cb: Function for tracking copy progress.
    :type progress_cb: Union[tqdm, Callable], optional
    :param progress_cb_type: Type of progress callback. Can be "number" or "size". Default is "size".
    :type progress_cb_type: Literal["number", "size"], optional
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely._utils import run_coroutine

        coroutine = sly.fs.copy_file_async('/home/admin/work/projects/example/1.png', '/home/admin/work/tests/2.png')
        run_coroutine(coroutine)
    """
    ensure_base_path(dst)
    async with aiofiles.open(dst, "wb") as out_f:
        async with aiofiles.open(src, "rb") as in_f:
            while True:
                chunk = await in_f.read(1024 * 1024)
                if not chunk:
                    break
                await out_f.write(chunk)
                if progress_cb is not None and progress_cb_type == "size":
                    progress_cb(len(chunk))
    if progress_cb is not None and progress_cb_type == "number":
        progress_cb(1)


async def get_file_hash_async(path: str) -> str:
    """
    Get hash from target file asynchronously.

    :param path: Target file path.
    :type path: str
    :returns: File hash
    :rtype: :class:`str`
    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely._utils import run_coroutine

        coroutine = sly.fs.get_file_hash_async('/home/admin/work/projects/examples/1.jpeg')
        hash = run_coroutine(coroutine)
    """
    async with aiofiles.open(path, "rb") as file:
        file_bytes = await file.read()
        return get_bytes_hash(file_bytes)


async def unpack_archive_async(
    archive_path: str, target_dir: str, remove_junk=True, is_split=False, chunk_size_mb: int = 50
) -> None:
    """
    Unpacks archive to the target directory, removes junk files and directories.
    To extract a split archive, you must pass the path to the first part in archive_path. Archive parts must be in the same directory. Format: archive_name.tar.001, archive_name.tar.002, etc. Works with tar and zip.
    You can adjust the size of the chunk to read from the file, while unpacking the file from parts.
    Be careful with this parameter, it can affect the performance of the function.

    :param archive_path: Path to the archive.
    :type archive_path: str
    :param target_dir: Path to the target directory.
    :type target_dir: str
    :param remove_junk: Remove junk files and directories. Default is True.
    :type remove_junk: bool
    :param is_split: Determines if the source archive is split into parts. If True, archive_path must be the path to the first part. Default is False.
    :type is_split: bool
    :param chunk_size_mb: Size of the chunk to read from the file. Default is 50Mb.
    :type chunk_size_mb: int
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely._utils import run_coroutine

        archive_path = '/home/admin/work/examples.tar'
        target_dir = '/home/admin/work/projects'

        coroutine = sly.fs.unpack_archive_async(archive_path, target_dir)
        run_coroutine(coroutine)
    """
    if is_split:
        chunk = chunk_size_mb * 1024 * 1024
        base_name = get_file_name(archive_path)
        dir_name = os.path.dirname(archive_path)
        if get_file_ext(base_name) in (".zip", ".tar"):
            ext = get_file_ext(base_name)
            base_name = get_file_name(base_name)
        else:
            ext = get_file_ext(archive_path)
        parts = sorted([f for f in os.listdir(dir_name) if f.startswith(base_name)])
        combined = os.path.join(dir_name, f"combined{ext}")

        async with aiofiles.open(combined, "wb") as output_file:
            for part in parts:
                part_path = os.path.join(dir_name, part)
                async with aiofiles.open(part_path, "rb") as input_file:
                    while True:
                        data = await input_file.read(chunk)
                        if not data:
                            break
                        await output_file.write(data)
        archive_path = combined

    loop = get_or_create_event_loop()
    await loop.run_in_executor(None, shutil.unpack_archive, archive_path, target_dir)
    if is_split:
        silent_remove(archive_path)
    if remove_junk:
        remove_junk_from_dir(target_dir)


async def touch_async(path: str) -> None:
    """
    Sets access and modification times for a file asynchronously.

    :param path: Target file path.
    :type path: str
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely._utils import run_coroutine

        coroutine = sly.fs.touch_async('/home/admin/work/projects/examples/1.jpeg')
        run_coroutine(coroutine)
    """
    ensure_base_path(path)
    async with aiofiles.open(path, "a"):
        loop = get_or_create_event_loop()
        await loop.run_in_executor(None, os.utime, path, None)


async def list_files_recursively_async(
    dir_path: str,
    valid_extensions: Optional[List[str]] = None,
    filter_fn: Optional[Callable[[str], bool]] = None,
    ignore_valid_extensions_case: bool = False,
) -> List[str]:
    """
    Recursively list files in the directory asynchronously.
    Returns list with all file paths.
    Can be filtered by valid extensions and filter function.

    :param dir_path: Target directory path.
    :type dir_path: str
    :param valid_extensions: List of valid extensions. Default is None.
    :type valid_extensions: Optional[List[str]]
    :param filter_fn: Filter function. Default is None.
    :type filter_fn: Optional[Callable[[str], bool]]
    :param ignore_valid_extensions_case: Ignore case when checking valid extensions. Default is False.
    :type ignore_valid_extensions_case: bool
    :returns: List of file paths
    :rtype: List[str]

    :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely._utils import run_coroutine

            dir_path = '/home/admin/work/projects/examples'

            coroutine = sly.fs.list_files_recursively_async(dir_path)
            files = run_coroutine(coroutine)
    """

    def sync_file_list():
        if valid_extensions and ignore_valid_extensions_case:
            valid_extensions_set = set(map(str.lower, valid_extensions))
        else:
            valid_extensions_set = set(valid_extensions) if valid_extensions else None

        files = []
        for dir_name, _, file_names in os.walk(dir_path):
            full_paths = [os.path.join(dir_name, filename) for filename in file_names]

            if valid_extensions_set:
                full_paths = [
                    fp
                    for fp in full_paths
                    if (
                        ext := (
                            os.path.splitext(fp)[1].lower()
                            if ignore_valid_extensions_case
                            else os.path.splitext(fp)[1]
                        )
                    )
                    in valid_extensions_set
                ]

            if filter_fn:
                full_paths = [fp for fp in full_paths if filter_fn(fp)]

            files.extend(full_paths)

        return files

    loop = get_or_create_event_loop()
    return await loop.run_in_executor(None, sync_file_list)


def get_file_offsets_batch_generator(
    archive_path: str,
    team_file_id: Optional[int] = None,
    filter_func: Optional[Callable] = None,
    output_format: Literal["dicts", "objects"] = "dicts",
    batch_size: int = OFFSETS_PKL_BATCH_SIZE,
) -> Generator[Union[List[Dict], List["BlobImageInfo"]], None, None]:
    """
    Extracts offset information for files from TAR archives and returns a generator that yields the information in batches.

    `team_file_id` may be None if it's not possible to obtain the ID at this moment.
    You can set the `team_file_id` later when uploading the file to Supervisely.

    :param archive_path: Local path to the archive
    :type archive_path: str
    :param team_file_id: ID of file in Team Files. Default is None.
                    `team_file_id` may be None if it's not possible to obtain the ID at this moment.
                    You can set the `team_file_id` later when uploading the file to Supervisely.
    :type team_file_id: Optional[int]
    :param filter_func: Function to filter files. The function should take a filename as input and return True if the file should be included.
    :type filter_func: Callable, optional
    :param output_format: Format of the output. Default is `dicts`.
                   `objects` - returns a list of BlobImageInfo objects.
                   `dicts` - returns a list of dictionaries.
    :type output_format: Literal["dicts", "objects"]
    :returns: Generator yielding batches of file information in the specified format.
    :rtype: Generator[Union[List[Dict], List[BlobImageInfo]]], None, None]

    :raises ValueError: If the archive type is not supported or contains compressed files
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        archive_path = '/home/admin/work/projects/examples.tar'
        file_infos = sly.fs.get_file_offsets_batch_generator(archive_path)
        for batch in file_infos:
            print(batch)

        # Output:
        # [
        #     {
        #         "title": "image1.jpg",
        #         "teamFileId": None,
        #         "sourceBlob": {
        #             "offsetStart": 0,
        #             "offsetEnd": 123456
        #         }
        #     },
        #     {
        #         "title": "image2.jpg",
        #         "teamFileId": None,
        #         "sourceBlob": {
        #             "offsetStart": 123456,
        #             "offsetEnd": 234567
        #         }
        #     }
        # ]
    """
    from supervisely.api.image_api import BlobImageInfo

    ext = Path(archive_path).suffix.lower()

    if ext == ".tar":
        if output_format == "dicts":
            yield from _process_tar_generator(
                tar_path=archive_path,
                team_file_id=team_file_id,
                filter_func=filter_func,
                batch_size=batch_size,
            )
        else:
            for batch in _process_tar_generator(
                tar_path=archive_path,
                team_file_id=team_file_id,
                filter_func=filter_func,
                batch_size=batch_size,
            ):
                blob_file_infos = [BlobImageInfo.from_dict(file_info) for file_info in batch]
                yield blob_file_infos
    else:
        raise ValueError(f"Unsupported archive type: {ext}. Only .tar are supported")


def _process_tar_generator(
    tar_path: str,
    team_file_id: Optional[int] = None,
    filter_func: Optional[Callable] = None,
    batch_size: int = OFFSETS_PKL_BATCH_SIZE,
) -> Generator[List[Dict], None, None]:
    """
    Processes a TAR archive and yields batches of offset information for files.

    :param tar_path: Path to the TAR archive
    :type tar_path: str
    :param team_file_id: ID of the team file, defaults to None
    :type team_file_id: Optional[int], optional
    :param filter_func: Function to filter files. The function should take a filename as input and return True if the file should be included.
    :type filter_func: Optional[Callable], optional
    :param batch_size: Number of files in each batch, defaults to 10000
    :type batch_size: int, optional
    :yield: Batches of dictionaries with file offset information
    :rtype: Generator[List[Dict], None, None]
    """
    from supervisely.api.api import ApiField

    with tarfile.open(tar_path, "r") as tar:
        batch = []
        processed_count = 0
        members = tar.getmembers()
        total_members_count = len(members)  # for logging

        logger.debug(f"Processing TAR archive with {total_members_count} members")

        for member in members:
            skip = not member.isfile()

            if filter_func and not filter_func(member.name):
                logger.debug(f"File '{member.name}' is skipped by filter function")
                skip = True

            if not skip:
                file_info = {
                    ApiField.TITLE: os.path.basename(member.name),
                    ApiField.TEAM_FILE_ID: team_file_id,
                    ApiField.SOURCE_BLOB: {
                        ApiField.OFFSET_START: member.offset_data,
                        ApiField.OFFSET_END: member.offset_data + member.size,
                    },
                }
                batch.append(file_info)

                # Yield batch when it reaches the specified size
                if len(batch) >= batch_size:
                    processed_count += len(batch)
                    logger.debug(
                        f"Yielding batch of {len(batch)} files, processed {processed_count} files so far"
                    )
                    yield batch
                    batch = []

        # Yield any remaining files in the last batch
        if batch:
            processed_count += len(batch)
            logger.debug(
                f"Yielding final batch of {len(batch)} files, processed {processed_count} files total"
            )
            yield batch


def save_blob_offsets_pkl(
    blob_file_path: str,
    output_dir: str,
    team_file_id: Optional[int] = None,
    filter_func: Optional[Callable] = None,
    batch_size: int = OFFSETS_PKL_BATCH_SIZE,
    replace: bool = False,
) -> str:
    """
    Processes blob file locally and creates a pickle file with offset information.

    :param blob_file_path: Path to the local blob file
    :type blob_file_path: str
    :param output_dir: Path to the output directory
    :type output_dir: str
    :param team_file_id: ID of file in Team Files. Default is None.
                    `team_file_id` may be None if it's not possible to obtain the ID at this moment.
                    You can set the `team_file_id` later when uploading the file to Supervisely.
    :type team_file_id: Optional[int]
    :param filter_func: Function to filter files. The function should take a filename as input and return True if the file should be included.
    :type filter_func: Callable, optional
    :param batch_size: Number of files to process in each batch, defaults to 10000
    :type batch_size: int, optional
    :param replace: If True, overwrite the existing file if it exists.
                    If False, skip processing if the file already exists and return its path.
                    Default is False.
    :type replace: bool
    :returns: Path to the output pickle file
    :rtype: str

    :Usage example:

        .. code-block:: python

            import supervisely as sly

            archive_path = '/path/to/examples.tar'
            output_dir = '/path/to/output'
            sly.fs.save_blob_offsets_pkl(archive_path, output_dir)
    """
    from supervisely.api.image_api import BlobImageInfo

    archive_name = Path(blob_file_path).stem
    output_path = os.path.join(output_dir, archive_name + OFFSETS_PKL_SUFFIX)

    if file_exists(output_path):
        logger.debug(f"Offsets file already exists: {output_path}")
        if replace:
            logger.debug(f"Replacing existing offsets file: {output_path}")
            silent_remove(output_path)
        else:
            logger.debug(f"Skipping processing, using existing offsets file: {output_path}")
            return output_path

    offsets_batch_generator = get_file_offsets_batch_generator(
        archive_path=blob_file_path,
        team_file_id=team_file_id,
        filter_func=filter_func,
        output_format="objects",
        batch_size=batch_size,
    )

    BlobImageInfo.dump_to_pickle(offsets_batch_generator, output_path)
    return output_path



# coding: utf-8
"""Functions for processing volumes"""


import os
from typing import List, Optional, Tuple, Union

import numpy as np
import pydicom
import SimpleITK as sitk
import stringcase
from trimesh import Trimesh

import supervisely.volume.nrrd_encoder as nrrd_encoder
from supervisely import logger
from supervisely.geometry.mask_3d import Mask3D
from supervisely.io.fs import get_file_ext, get_file_name, list_files_recursively
from supervisely.volume.stl_converter import matrix_from_nrrd_header

# Do NOT use directly for extension validation. Use is_valid_ext() /  has_valid_ext() below instead.
ALLOWED_VOLUME_EXTENSIONS = [".nrrd", ".dcm"]


class UnsupportedVolumeFormat(Exception):
    pass


def get_extension(path: str):
    """
    Get extension for given path.

    :param path: Path to volume.
    :type path: str
    :return: Path extension
    :rtype: str
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        path = "src/upload/folder/CTACardio.nrrd"
        ext = sly.volume.get_extension(path=path) # .nrrd
    """

    # magic.from_file("path", mime=True)
    # for nrrd:
    # application/octet-stream
    # for nifti(nii):
    # application/octet-stream
    # for dicom:
    # "application/dicom"

    ext = get_file_ext(path)
    if ext in ALLOWED_VOLUME_EXTENSIONS:
        return ext

    # case when dicom file does not have an extension
    import magic

    mime = magic.from_file(path, mime=True)
    if mime == "application/dicom":
        return ".dcm"
    return None


def is_valid_ext(ext: str) -> bool:
    """
    Checks if given extension is supported.

    :param ext: Volume file extension.
    :type ext: str
    :return: True if extensions is in the list of supported extensions else False
    :rtype: :class:`bool`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        sly.volume.is_valid_ext(".nrrd")  # True
        sly.volume.is_valid_ext(".mp4") # False
    """

    if type(ext) is not str:
        return False

    return ext.lower() in ALLOWED_VOLUME_EXTENSIONS


def has_valid_ext(path: str) -> bool:
    """
    Checks if Volume file from given path has supported extension.

    :param path: Path to volume file.
    :type path: str
    :return: True if Volume file has supported extension else False
    :rtype: :class:`bool`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        volume_path = "/home/admin/work/volumes/vol_01.nrrd"
        sly.volume.has_valid_ext(volume_path) # True
    """

    return is_valid_ext(get_extension(path))


def validate_format(path: str):
    """
    Raise error if Volume file from given path couldn't be read or file extension is not supported.

    :param path: Path to Volume file.
    :type path: str
    :raises: :class:`UnsupportedVolumeFormat` if Volume file from given path couldn't be read or file extension is not supported.
    :return: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        volume_path = "/home/admin/work/volumes/vol_01.mp4"
        sly.volume.validate_format(volume_path)
        # File /home/admin/work/volumes/vol_01.mp4 has unsupported volume extension. Supported extensions: [".nrrd", ".dcm"].
    """

    if not has_valid_ext(path):
        raise UnsupportedVolumeFormat(
            f"File {path} has unsupported volume extension. Supported extensions: {ALLOWED_VOLUME_EXTENSIONS}"
        )


def is_valid_format(path: str) -> bool:
    """
    Checks if a given file has a supported format.
    :param path: Path to file.
    :type path: str
    :return: True if file format in list of supported Volume formats, False - in otherwise
    :rtype: :class:`bool`
    :Usage example:
         .. code-block:: python
            import supervisely as sly
            sly.volume.is_valid_format('/volumes/dcm01.dcm') # True
            sly.volume.is_valid_format('/volumes/nrrd.py') # False
    """

    try:
        validate_format(path)
        return True
    except UnsupportedVolumeFormat:
        return False


def rescale_slope_intercept(value: float, slope: float, intercept: float) -> float:
    """
    Rescale intensity value using the given slope and intercept.

    :param value: The intensity value to be rescaled.
    :type value: float
    :param slope: The slope for rescaling.
    :type slope: float
    :param intercept: The intercept for rescaling.
    :type intercept: float
    :return: The rescaled intensity value.
    :rtype: float

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        meta["intensity"]["min"] = sly.volume.volume.rescale_slope_intercept(
            meta["intensity"]["min"],
            meta["rescaleSlope"],
            meta["rescaleIntercept"],
        )
    """

    return value * slope + intercept


def normalize_volume_meta(meta: dict) -> dict:
    """
    Normalize volume metadata.

    :param meta: Metadata of the volume.
    :type meta: dict
    :return: Normalized volume metadata.
    :rtype: dict

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        normalized_meta = sly.volume.volume.volume.normalize_volume_meta(volume_meta)

        print(normalized_meta)
        # Output:
        # {
        #     'ACS': 'RAS',
        #     'channelsCount': 1,
        #     'dimensionsIJK': {'x': 512, 'y': 512, 'z': 139},
        #     'directions': (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0),
        #     'intensity': {'max': 3071.0, 'min': -3024.0},
        #     'origin': (-194.238403081894, -217.5384061336518, -347.7500000000001),
        #     'rescaleIntercept': 0,
        #     'rescaleSlope': 1,
        #     'spacing': (0.7617189884185793, 0.7617189884185793, 2.5),
        #     'windowCenter': 23.5,
        #     'windowWidth': 6095.0
        # }
    """

    meta["intensity"]["min"] = rescale_slope_intercept(
        meta["intensity"]["min"],
        meta["rescaleSlope"],
        meta["rescaleIntercept"],
    )

    meta["intensity"]["max"] = rescale_slope_intercept(
        meta["intensity"]["max"],
        meta["rescaleSlope"],
        meta["rescaleIntercept"],
    )

    if "windowWidth" not in meta:
        meta["windowWidth"] = meta["intensity"]["max"] - meta["intensity"]["min"]

    if "windowCenter" not in meta:
        meta["windowCenter"] = meta["intensity"]["min"] + meta["windowWidth"] / 2

    return meta


def read_dicom_serie_volume_np(paths: List[str], anonymize=True) -> Tuple[np.ndarray, dict]:
    """
    Read DICOM series volumes with given paths.

    :param paths: Paths to DICOM volume files.
    :type paths: List[str]
    :param anonymize: Specify whether to hide PatientID and PatientName fields.
    :type anonymize: bool
    :return: Volume data in NumPy array format and dictionary with metadata
    :rtype: Tuple[np.ndarray, dict]
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        volume_path = ["/home/admin/work/volumes/vol_01.nrrd"]
        volume_np, meta = sly.volume.read_dicom_serie_volume_np(volume_path)
    """

    import SimpleITK as sitk

    sitk_volume, meta = read_dicom_serie_volume(paths, anonymize=anonymize)
    # for debug:
    # sitk.WriteImage(sitk_volume, "/work/output/sitk.nrrd", useCompression=False, compressionLevel=9)
    # with open("/work/output/test.nrrd", "wb") as file:
    #     file.write(b)
    volume_np = sitk.GetArrayFromImage(sitk_volume)
    volume_np = np.transpose(volume_np, (2, 1, 0))
    return volume_np, meta


_anonymize_tags = ["PatientID", "PatientName"]
_default_dicom_tags = [
    "SeriesInstanceUID",
    "Modality",
    "WindowCenter",
    "WindowWidth",
    "RescaleIntercept",
    "RescaleSlope",
    "PhotometricInterpretation",
]
_default_dicom_tags.extend(_anonymize_tags)

_photometricInterpretationRGB = set(
    [
        "RGB",
        "PALETTE COLOR",
        "YBR_FULL",
        "YBR_FULL_422",
        "YBR_PARTIAL_422",
        "YBR_PARTIAL_420",
        "YBR_RCT",
    ]
)


def read_dicom_tags(
    path: str,
    allowed_keys: Union[None, List[str]] = _default_dicom_tags,
    anonymize: bool = True,
):
    """
    Read DICOM tags from a DICOM file.

    :param path: Path to the DICOM file.
    :type path: str
    :param allowed_keys: List of allowed DICOM keywords to be extracted. Default is None, which means all keywords are allowed.
    :type allowed_keys: Union[None, List[str]], optional
    :param anonymize: Flag to indicate whether to anonymize certain tags or not.
    :type anonymize: bool, optional
    :return: Dictionary containing the extracted DICOM tags.
    :rtype: dict

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        path = "src/upload/Dicom_files/nnn.dcm"
        dicom_tags = sly.volume.read_dicom_tags(path=path)
    """

    import SimpleITK as sitk

    reader = sitk.ImageFileReader()
    reader.SetFileName(path)
    reader.LoadPrivateTagsOn()
    reader.ReadImageInformation()

    vol_info = {}
    for k in reader.GetMetaDataKeys():
        v = reader.GetMetaData(k)
        tag = pydicom.tag.Tag(k.split("|")[0], k.split("|")[1])
        keyword = pydicom.datadict.keyword_for_tag(tag)
        if allowed_keys is not None and keyword not in allowed_keys:
            continue
        if anonymize is True and keyword in _anonymize_tags:
            v = "anonymized"
        keyword = stringcase.camelcase(keyword)
        vol_info[keyword] = v
        if keyword in [
            "windowCenter",
            "windowWidth",
            "rescaleIntercept",
            "rescaleSlope",
        ]:
            vol_info[keyword] = float(vol_info[keyword].split("\\")[0])
        elif keyword == "photometricInterpretation" and v in _photometricInterpretationRGB:
            vol_info["channelsCount"] = 3
    return vol_info


def encode(volume_np: np.ndarray, volume_meta: dict) -> bytes:
    """
    Encodes a volume from NumPy format into a NRRD format.

    :param volume_np: NumPy array representing the volume data.
    :type volume_np: np.ndarray
    :param volume_meta: Metadata of the volume.
    :type volume_meta: dict

    :return: Encoded volume data in bytes.
    :rtype: bytes

    :Usage example:

     .. code-block:: python

        import numpy as np
        import supervisely as sly

        volume_np = np.random.rand(256, 256, 256)
        volume_meta = {
            "ACS": "RAS",
            "channelsCount": 1,
            "dimensionsIJK": { "x": 512, "y": 512, "z": 139 },
            "directions": (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0),
            "intensity": { "max": 3071.0, "min": -3024.0 },
            "origin": (-194.238403081894, -217.5384061336518, -347.7500000000001),
            "rescaleIntercept": 0,
            "rescaleSlope": 1,
            "spacing": (0.7617189884185793, 0.7617189884185793, 2.5),
            "windowCenter": 23.5,
            "windowWidth": 6095.0
        }

        encoded_volume = sly.volume.encode(volume_np, volume_meta)
    """

    directions = np.array(volume_meta["directions"]).reshape(3, 3)
    directions *= volume_meta["spacing"]

    volume_bytes = nrrd_encoder.encode(
        volume_np,
        header={
            "encoding": "gzip",
            # "space": "left-posterior-superior",
            "space": "right-anterior-superior",
            "space directions": directions.T.tolist(),
            "space origin": volume_meta["origin"],
        },
        compression_level=1,
    )

    # with open("/work/output/test.nrrd", "wb") as file:
    #     file.write(volume_bytes)

    return volume_bytes


def inspect_dicom_series(root_dir: str, logging: bool = True) -> dict:
    """
    Search for DICOM series in the directory and its subdirectories.
    If several series with the same UID are found in the directory, then the series are numbered in the format: "series_uid_01", "series_uid_02", etc.

    :param root_dir: Directory path with volumes.
    :type root_dir: str
    :param logging: Specify whether to print logging messages.
    :type logging: bool
    :return: Dictionary with DICOM volumes IDs and corresponding file names.
    :rtype: dict
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        path = "src/upload/Dicom_files/"
        series_infos = sly.volume.inspect_dicom_series(root_dir=path)
    """
    import SimpleITK as sitk

    found_series = {}
    for d in os.walk(root_dir):
        dir = d[0]
        reader = sitk.ImageSeriesReader()
        sitk.ProcessObject_SetGlobalWarningDisplay(False)
        series_found = reader.GetGDCMSeriesIDs(dir)
        sitk.ProcessObject_SetGlobalWarningDisplay(True)
        if logging:
            logger.info(f"Found {len(series_found)} series in directory {dir}")
        for serie in series_found:
            dicom_names = reader.GetGDCMSeriesFileNames(dir, serie)
            new_key = serie
            new_suffix = 1
            while new_key in found_series:
                new_key = "{}_{:02d}".format(serie, new_suffix)
                new_suffix += 1
            found_series[new_key] = dicom_names
    if logging:
        logger.info(f"Total {len(found_series)} series in directory {root_dir}")
    return found_series


def _sitk_image_orient_ras(sitk_volume: sitk.Image) -> sitk.Image:
    import SimpleITK as sitk

    if sitk_volume.GetDimension() == 4 and sitk_volume.GetSize()[3] == 1:
        sitk_volume = sitk_volume[:, :, :, 0]

    sitk_volume = sitk.DICOMOrient(sitk_volume, "RAS")
    # RAS reorient image using filter
    # orientation_filter = sitk.DICOMOrientImageFilter()
    # orientation_filter.SetDesiredCoordinateOrientation("RAS")
    # sitk_volume = orientation_filter.Execute(sitk_volume)

    # https://discourse.itk.org/t/getdirection-and-getorigin-for-simpleitk-c-implementation/3472/8
    origin = list(sitk_volume.GetOrigin())
    directions = list(sitk_volume.GetDirection())
    origin[0] *= -1
    origin[1] *= -1
    directions[0] *= -1
    directions[1] *= -1
    directions[3] *= -1
    directions[4] *= -1
    directions[6] *= -1
    directions[7] *= -1
    sitk_volume.SetOrigin(origin)
    sitk_volume.SetDirection(directions)
    return sitk_volume


def read_dicom_serie_volume(paths: List[str], anonymize: bool = True) -> Tuple[sitk.Image, dict]:
    """
    Read DICOM series volumes with given paths.

    :param paths: Paths to DICOM volume files.
    :type paths: List[str]
    :param anonymize: Specify whether to hide PatientID and PatientName fields.
    :type anonymize: bool
    :return: Volume data in SimpleITK.Image format and dictionary with metadata.
    :rtype: Tuple[SimpleITK.Image, dict]
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        paths = ["/home/admin/work/volumes/vol_01.nrrd"]
        sitk_volume, meta = sly.volume.read_dicom_serie_volume(paths)
    """

    import SimpleITK as sitk

    reader = sitk.ImageSeriesReader()
    reader.SetFileNames(paths)
    sitk_volume = reader.Execute()
    sitk_volume = _sitk_image_orient_ras(sitk_volume)
    dicom_tags = read_dicom_tags(paths[0], anonymize=anonymize)

    f_min_max = sitk.MinimumMaximumImageFilter()
    f_min_max.Execute(sitk_volume)
    meta = get_meta(
        sitk_volume.GetSize(),
        f_min_max.GetMinimum(),
        f_min_max.GetMaximum(),
        sitk_volume.GetSpacing(),
        sitk_volume.GetOrigin(),
        sitk_volume.GetDirection(),
        dicom_tags,
    )
    return sitk_volume, meta


def compose_ijk_2_world_mat(meta: dict) -> np.ndarray:
    """
    Transform 4x4 matrix from voxels to world coordinates.

    :param meta: Volume metadata.
    :type meta: dict
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        mat = sly.volume.volume.compose_ijk_2_world_mat(volume_meta)

        # Output:
        # [
        #     [   0.76171899    0.            0.         -194.23840308]
        #     [   0.            0.76171899    0.         -217.53840613]
        #     [   0.            0.            2.5        -347.75      ]
        #     [   0.            0.            0.            1.        ]
        # ]
    """

    try:
        spacing = meta["spacing"]
        origin = meta["origin"]
        directions = meta["directions"]
    except KeyError as e:
        raise IOError(
            f"Need the meta '{e}'' field to determine the mapping from voxels to world coordinates."
        )

    mat = np.eye(4)
    mat[:3, :3] = (np.array(directions).reshape(3, 3) * spacing).T
    mat[:3, 3] = origin
    return mat


def world_2_ijk_mat(ijk_2_world) -> np.ndarray:
    """
    Transform 4x4 matrix from world to voxels coordinates.

    :param ijk_2_world: 4x4 matrix.
    :type ijk_2_world: np.ndarray
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        mat = sly.volume.volume.world_2_ijk_mat(world_mat)

        # Output:
        # [
        #     [  1.3128201    0.           0.         255.00008013]
        #     [  0.           1.3128201    0.         285.58879251]
        #     [  0.           0.           0.4        139.1       ]
        #     [  0.           0.           0.           1.        ]
        # ]
    """

    return np.linalg.inv(ijk_2_world)


def get_meta(
    sitk_shape: tuple,
    min_intensity: float,
    max_intensity: float,
    spacing: tuple,
    origin: tuple,
    directions: tuple,
    dicom_tags: dict = {},
) -> dict:
    """
    Get normalized meta-data for a volume.

    :param sitk_shape: Tuple representing the shape of the volume in (x, y, z) dimensions.
    :type sitk_shape: tuple
    :param min_intensity: Minimum intensity value in the volume.
    :type min_intensity: float
    :param max_intensity: Maximum intensity value in the volume.
    :type max_intensity: float
    :param spacing: Tuple representing the spacing between voxels in (x, y, z) dimensions.
    :type spacing: tuple
    :param origin: Tuple representing the origin of the volume in (x, y, z) dimensions.
    :type origin: tuple
    :param directions: Tuple representing the direction matrix of the volume.
    :type directions: tuple
    :param dicom_tags: Dictionary containing additional DICOM tags for the volume meta-data.
    :type dicom_tags: dict, optional
    :return: Dictionary containing the normalized meta-data for the volume.
    :rtype: dict

    :Usage example:

     .. code-block:: python

        import SimpleITK as sitk
        import supervisely as sly

        path = "/home/admin/work/volumes/vol_01.nrrd"

        reader = sitk.ImageSeriesReader()
        reader.SetFileNames(path)
        sitk_volume = reader.Execute()
        sitk_volume = _sitk_image_orient_ras(sitk_volume)
        dicom_tags = read_dicom_tags(paths[0], anonymize=anonymize)

        f_min_max = sitk.MinimumMaximumImageFilter()
        f_min_max.Execute(sitk_volume)
        meta = get_meta(
            sitk_volume.GetSize(),
            f_min_max.GetMinimum(),
            f_min_max.GetMaximum(),
            sitk_volume.GetSpacing(),
            sitk_volume.GetOrigin(),
            sitk_volume.GetDirection(),
            dicom_tags,
        )
    """

    # x = 1 - sagittal
    # y = 1 - coronal
    # z = 1 - axial
    volume_meta = normalize_volume_meta(
        {
            **dicom_tags,
            "channelsCount": 1,
            "rescaleSlope": 1,
            "rescaleIntercept": 0,
            "intensity": {
                "min": min_intensity,
                "max": max_intensity,
            },
            "dimensionsIJK": {
                "x": sitk_shape[0],
                "y": sitk_shape[1],
                "z": sitk_shape[2],
            },
            "ACS": "RAS",
            # instead of IJK2WorldMatrix field
            "spacing": spacing,
            "origin": origin,
            "directions": directions,
        }
    )
    return volume_meta


def inspect_nrrd_series(root_dir: str, logging: bool = True) -> List[str]:
    """
    Inspect a directory for NRRD series by recursively listing files with the ".nrrd" extension and returns a list of NRRD file paths found in the directory.

    :param root_dir: Directory to inspect for NRRD series.
    :type root_dir: str
    :param logging: Specify whether to print logging messages.
    :type logging: bool
    :return: List of NRRD file paths found in the given directory.
    :rtype: List[str]
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        path = "/home/admin/work/volumes/"
        nrrd_paths = sly.volume.inspect_nrrd_series(root_dir=path)
    """

    nrrd_paths = list_files_recursively(root_dir, [".nrrd"])
    if logging:
        logger.info(f"Total {len(nrrd_paths)} NRRD series in directory {root_dir}")
    return nrrd_paths


def read_nrrd_serie_volume(path: str) -> Tuple[sitk.Image, dict]:
    """
    Read NRRD volume with given path.

    :param path: Path to NRRD volume files.
    :type path: List[str]
    :return: Volume data in SimpleITK.Image format and dictionary with metadata.
    :rtype: Tuple[SimpleITK.Image, dict]
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        path = "/home/admin/work/volumes/vol_01.nrrd"
        sitk_volume, meta = sly.volume.read_nrrd_serie_volume(path)
    """

    import SimpleITK as sitk

    # find custom NRRD loader in gitlab supervisely_py/-/blob/feature/import-volumes/plugins/import/volumes/src/loaders/nrrd.py
    reader = sitk.ImageFileReader()
    # reader.SetImageIO("NrrdImageIO")
    reader.SetFileName(path)
    sitk_volume = reader.Execute()

    sitk_volume = _sitk_image_orient_ras(sitk_volume)
    f_min_max = sitk.MinimumMaximumImageFilter()
    f_min_max.Execute(sitk_volume)
    meta = get_meta(
        sitk_volume.GetSize(),
        f_min_max.GetMinimum(),
        f_min_max.GetMaximum(),
        sitk_volume.GetSpacing(),
        sitk_volume.GetOrigin(),
        sitk_volume.GetDirection(),
        {},
    )
    return sitk_volume, meta


def read_nrrd_serie_volume_np(paths: str) -> Tuple[np.ndarray, dict]:
    """
    Read NRRD volume with given path.

    :param paths: Path to NRRD volume file.
    :type paths: str
    :return: Volume data in NumPy array format and dictionary with metadata.
    :rtype: Tuple[np.ndarray, dict]
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        path = "/home/admin/work/volumes/vol_01.nrrd"
        np_volume, meta = sly.volume.read_nrrd_serie_volume_np(path)
    """

    import SimpleITK as sitk

    sitk_volume, meta = read_nrrd_serie_volume(paths)
    volume_np = sitk.GetArrayFromImage(sitk_volume)
    volume_np = np.transpose(volume_np, (2, 1, 0))
    return volume_np, meta


def convert_nifti_to_nrrd(path: str) -> Tuple[np.ndarray, dict]:
    """Convert NIFTI volume to NRRD format.
    Volume automatically reordered to RAS orientation as closest to canonical.

    :param path: Path to NIFTI volume file.
    :type path: str
    :return: Volume data in NumPy array format and dictionary with metadata (NRRD header).
    :rtype: Tuple[np.ndarray, dict]
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        path = "/home/admin/work/volumes/vol_01.nii"
        data, header = sly.volume.convert_nifti_to_nrrd(path)
    """

    import nibabel as nib  # pylint: disable=import-error

    nifti = nib.load(path)
    reordered_to_ras_nifti = nib.as_closest_canonical(nifti)
    data = reordered_to_ras_nifti.get_fdata()
    affine = reordered_to_ras_nifti.affine
    orientation = nib.aff2axcodes(affine)
    header = {
        "space": "".join(orientation),
        "space directions": affine.tolist(),
        "sizes": data.shape,
        "type": str(data.dtype),
        "dimension": len(data.shape),
    }
    return data, header


def convert_3d_nifti_to_nrrd(path: str) -> Tuple[np.ndarray, dict]:
    """Convert 3D NIFTI volume to NRRD format.
    Volume automatically reordered to RAS orientation as closest to canonical.

    :param path: Path to NIFTI volume file.
    :type path: str
    :return: Volume data in NumPy array format and dictionary with metadata (NRRD header).
    :rtype: Tuple[np.ndarray, dict]
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        path = "/home/admin/work/volumes/vol_01.nii"
        data, header = sly.volume.convert_nifti_to_nrrd(path)
    """
    import SimpleITK as sitk

    nifti_image = sitk.ReadImage(path)
    nifti_image = _sitk_image_orient_ras(nifti_image)
    data = sitk.GetArrayFromImage(nifti_image)
    data = np.transpose(data, (2, 1, 0))

    direction = np.array(nifti_image.GetDirection()).reshape(3, 3)
    spacing = np.array(nifti_image.GetSpacing())
    origin = np.array(nifti_image.GetOrigin())

    space_directions = (direction.T * spacing[:, None]).tolist()

    header = {
        "dimension": 3,
        "space": "right-anterior-superior",
        "sizes": list(data.shape),
        "space directions": space_directions,
        "endian": "little",
        "encoding": "gzip",
        "space origin": origin,
    }
    return data, header


def is_nifti_file(path: str) -> bool:
    """Check if the file is a NIFTI file.

    :param filepath: Path to the file.
    :type filepath: str
    :return: True if the file is a NIFTI file, False otherwise.
    :rtype: bool
    """
    try:
        import nibabel as nib
    except ImportError:
        raise ImportError(
            "No module named nibabel. Please make sure that module is installed from pip and try again."
        )

    try:
        nib.load(path)
        return True
    except nib.filebasedimages.ImageFileError:
        return False


def convert_3d_geometry_to_mesh(
    geometry: Mask3D,
    spacing: tuple = (1.0, 1.0, 1.0),
    level: float = 0.5,
    apply_decimation: bool = False,
    decimation_fraction: float = 0.5,
    volume_meta: Optional[dict] = None,
) -> Trimesh:
    """
    Converts a 3D geometry (Mask3D) to a Trimesh mesh.

    :param geometry: The 3D geometry to convert.
    :type geometry: supervisely.geometry.mask_3d.Mask3D
    :param spacing: Voxel spacing in (x, y, z).
    :type spacing: tuple
    :param level: Isosurface value for marching cubes. Default is 0.5.
    :type level: float
    :param apply_decimation: Whether to simplify the mesh. Default is False.
    :type apply_decimation: bool
    :param decimation_fraction: Fraction of faces to keep if decimation is applied. Default is 0.5.
    :type decimation_fraction: float
    :param volume_meta: Metadata of the volume. Used for mesh alignment if geometry lacks specific fields. Default is None.
    :type volume_meta: dict, optional
    :return: The resulting Trimesh mesh.
    :rtype: trimesh.Trimesh

    :Usage example:

        .. code-block:: python

            volume_header = nrrd.read_header("path/to/volume.nrrd")
            mask3d = Mask3D.create_from_file("path/to/mask3d")
            mesh = convert_3d_geometry_to_mesh(mask3d, spacing=(1.0, 1.0, 1.0), level=0.7, apply_decimation=True, volume_meta=volume_header)
    """
    from skimage import measure

    if volume_meta is None:
        volume_meta = {}

    space_directions = geometry.space_directions or volume_meta.get("space directions")
    space_origin = geometry.space_origin or volume_meta.get("space origin")

    verts, faces, normals, _ = measure.marching_cubes(geometry.data, level=level, spacing=spacing)
    mesh = Trimesh(vertices=verts, faces=faces, vertex_normals=normals, process=False)

    if apply_decimation and 0 < decimation_fraction < 1:
        mesh = mesh.simplify_quadric_decimation(int(len(mesh.faces) * decimation_fraction))

    if space_directions is not None and space_origin is not None:
        header = {
            "space directions": space_directions,
            "space origin": space_origin,
        }
        align_mesh_to_volume(mesh, header)

    # flip x and y axes to match initial mask orientation
    mesh.apply_transform(np.diag([-1, -1, 1, 1]))

    mesh.fix_normals()

    return mesh


def export_3d_as_mesh(geometry: Mask3D, output_path: str, **kwargs):
    """
    Exports the 3D mesh representation of the object to a file in either STL or OBJ format.

    :param geometry: The 3D geometry to be exported.
    :type geometry: supervisely.geometry.mask_3d.Mask3D
    :param output_path: The path to the output file. Must have a ".stl" or ".obj" extension.
    :type output_path: str
    :param kwargs: Additional keyword arguments for mesh generation. Supported keys:
        - spacing (tuple): Voxel spacing in (x, y, z). By default the value will be taken from geometry meta.
        - level (float): Isosurface value for marching cubes. Default is 0.5.
        - apply_decimation (bool): Whether to simplify the mesh. Default is False.
        - decimation_fraction (float): Fraction of faces to keep if decimation is applied. Default is 0.5.
        - volume_meta (dict): Metadata of the volume. Used for mesh alignment if geometry lacks specific fields. Default is None.
    :return: None

    :Usage example:

        .. code-block:: python

        mask3d_path = "path/to/mask3d"
        mask3d = Mask3D.create_from_file(mask3d_path)

        mask3d.export_3d_as_mesh(mask3d, "output.stl", spacing=(1.0, 1.0, 1.0), level=0.7, apply_decimation=True)
    """

    if get_file_ext(output_path).lower() not in [".stl", ".obj"]:
        raise ValueError('File extension must be either ".stl" or ".obj"')

    mesh = convert_3d_geometry_to_mesh(geometry, **kwargs)
    mesh.export(output_path)


def align_mesh_to_volume(mesh: Trimesh, volume_header: dict) -> None:
    """
    Transforms the given mesh in-place using spatial information from an NRRD header.
    The mesh will be tranformed to match the coordinate system defined in the header.

    :param mesh: The mesh object to be transformed. The transformation is applied in-place.
    :type mesh: Trimesh
    :param volume_header: The NRRD header containing spatial metadata, including "space directions",
        "space origin", and "space". Field "space" should be in the format of
        "right-anterior-superior", "left-anterior-superior", etc.
    :type volume_header: dict
    :returns: None
    :rtype: None
    """
    from supervisely.geometry.constants import SPACE_ORIGIN
    from supervisely.geometry.mask_3d import PointVolume

    if isinstance(volume_header["space origin"], PointVolume):
        volume_header["space origin"] = volume_header["space origin"].to_json()[SPACE_ORIGIN]
    transform_mat = matrix_from_nrrd_header(volume_header)
    mesh.apply_transform(transform_mat)
